//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ------------------------------------------------------------------------

TP.lang.Object.defineSubtype('test.GUIDriver');

//  ------------------------------------------------------------------------
//  Type Attributes
//  ------------------------------------------------------------------------

/**
 * Whether or not the keymaps that this type manages have been set up.
 * @type {Boolean}
 */
TP.test.GUIDriver.Type.defineAttribute('$mapsAreSetup');

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

/**
 * The Window context that this GUI driver is currently operating in. This is
 * the default context used to find elements, etc.
 * @type {TP.core.Window}
 */
TP.test.GUIDriver.Inst.defineAttribute('windowContext');

/**
 * An object that will provide an API to manage Promises for this driver. When
 * executing in the test harness, this will typically be the currently executing
 * test case.
 * @type {Object}
 */
TP.test.GUIDriver.Inst.defineAttribute('promiseProvider');

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('init',
function(windowContext) {

    /**
     * @method init
     * @summary Initialize the instance.
     * @param {TP.core.Window} windowContext The initial window context to use
     *     to resolve GUI element references, etc.
     * @returns {TP.test.GUIDriver} A new instance.
     */

    this.callNextMethod();

    this.set('windowContext', windowContext);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('fetchResource',
function(aURI, aRequest) {

    /**
     * @method fetchResource
     * @summary Fetches the resource at the end of the URI. The result will be
     *     available as the value of the returned promise.
     * @param {TP.uri.URI} The URI to fetch the resource for.
     * @exception TP.sig.InvalidURI
     * @returns {Promise} A Promise which completes when the resource is
     *     available.
     */

    if (!TP.isKindOf(aURI, TP.uri.URI)) {
        return this.raise('TP.sig.InvalidURI');
    }

    //  A Promise that will await the 'RequestSucceeded' signal and resolve the
    //  Promise with the result.
    return Promise.construct(
                function(resolver, rejector) {
                    var subrequest;

                    subrequest = TP.request(aRequest);

                    subrequest.defineHandler(
                        'RequestSucceeded',
                        function(aResponse) {
                            //  We succeeded - call the Promise's
                            //  resolver.
                            resolver(aResponse.getResult());
                        });

                    subrequest.defineHandler(
                        'RequestFailed',
                        function(aResponse) {
                            //  We failed - call the Promise's rejector.
                            rejector(aResponse.getResult());
                        });

                    return aURI.getResource(subrequest);
                });
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('getCurrentNativeDocument',
function() {

    /**
     * @method getCurrentNativeDocument
     * @summary Returns the native Document object associated with the current
     *     window context.
     * @returns {Document} The Document of the current window context.
     */

    var context;

    context = this.get('windowContext');

    return context.getNativeDocument();
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('getFocusedElement',
function() {

    /**
     * @method getFocusedElement
     * @summary Returns the Element that currently has focus in the current
     *     window context.
     * @returns {Element} The focused Element in the current window context.
     */

    var context;

    context = this.get('windowContext');

    return TP.documentGetFocusedElement(context.getNativeDocument());
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('setBodyContent',
function(aURI, aWindow) {

    /**
     * @method setBodyContent
     * @summary Sets the 'body' content of the (X)HTML document in the supplied
     *     Window to the content found at the end of the URI.
     * @param {TP.uri.URI} The URI to fetch content from.
     * @param {TP.core.Window} The Window to use the body of to load the content
     *     into. This will default to the current UI canvas.
     * @exception TP.sig.InvalidURI
     * @returns {Promise} A Promise which completes when the resource is
     *     available.
     */

    if (!TP.isKindOf(aURI, TP.uri.URI)) {
        return this.raise('TP.sig.InvalidURI');
    }

    //  Fetch the result and then set the Window's body to the result.
    return this.fetchResource(aURI, TP.hc('resultType', TP.DOM)).then(
        function(result) {
            var tpWin,
                tpDoc,
                tpBody;

            if (TP.isValid(aWindow)) {
                tpWin = aWindow;
            } else {
                tpWin = this.get('windowContext');
            }

            tpDoc = tpWin.getDocument();
            tpBody = tpDoc.getBody();

            return tpBody.setContent(result);
        }.bind(this),
        function(error) {
            TP.sys.logTest('Couldn\'t get resource: ' + aURI.getLocation(),
                            TP.Log.ERROR);
        });
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('setLocation',
function(aURI, aWindow, aRequest) {

    /**
     * @method setLocation
     * @summary Sets the location of the supplied Window to the content found
     *     at the end of the URI.
     * @param {TP.uri.URI} The URI to fetch content from.
     * @param {TP.core.Window} The Window to load the content into. This will
     *     default to the current UI canvas.
     * @param {TP.sig.Request} [aRequest] A request containing control
     *     parameters.
     * @exception TP.sig.InvalidURI
     * @async
     * @returns {Promise} A Promise which completes when the URI is *completely*
     *     loaded into the supplied Window.
     */

    var tpWin,
        request;

    if (!TP.isKindOf(aURI, TP.uri.URI)) {
        return this.raise('TP.sig.InvalidURI');
    }

    if (TP.isValid(aWindow)) {
        tpWin = aWindow;
    } else {
        tpWin = this.get('windowContext');
    }

    request = TP.request(aRequest);

    return tpWin.setLocation(aURI, request);
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('showTestGUI',
function() {

    /**
     * @method showTestGUI
     * @summary Shows the 'test GUI'. The process by which it does this is
     *     environment dependent (i.e. what tool we're running in, etc).
     * @returns {TP.test.GUIDriver} The receiver.
     */

    /*
       TODO: The TDC is no longer with us - remove this when appropriate
    if (TP.sys.cfg('tibet.tdc') === true) {
        TP.elementHide(TP.byId('UIBOOT', top, false));
        TP.elementShow(TP.byId('UIROOT', top, false));
    } else if (TP.sys.cfg('lama.enabled') === true) {
        //  TODO: Make the appropriate GUI window show for the Lama, etc.
        void 0;
    }
    */

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('showTestLog',
function() {

    /**
     * @method showTestLog
     * @summary Shows the 'test log'. The process by which it does this is
     *     environment dependent (i.e. what tool we're running in, etc).
     * @returns {TP.test.GUIDriver} The receiver.
     */

    /*
       TODO: The TDC is no longer with us - remove this when appropriate
    if (TP.sys.cfg('tibet.tdc') === true) {
        TP.elementHide(TP.byId('UIROOT', top, false));
        TP.elementShow(TP.byId('UIBOOT', top, false));
    } else if (TP.sys.cfg('lama.enabled') === true) {
        //  TODO: Make the appropriate log window show for the Lama, etc.
        void 0;
    }
    */

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('constructSequence',
function() {

    /**
     * @method constructSequence
     * @summary Returns a new GUI sequence used to script actions.
     * @returns {TP.test.GUISequence} A new GUI sequence.
     */

    return TP.test.GUISequence.construct(this);
});

//  ------------------------------------------------------------------------

TP.test.GUIDriver.Inst.defineMethod('takeScreenshotOf',
function(aNode) {

    /**
     * @method takeScreenshotOf
     * @summary Take a screenshot of either the whole window of the currently
     *     executing window context or some portion thereof as determined by the
     *     supplied Node.
     * @description If a Document node is supplied to this method, the whole
     *     Document will be snapshotted. If an Element is supplied, only that
     *     Element will be snapshotted. Note also that this method currently
     *     only works when this code is being executed in the Headless
     *     environment.
     * @param {Node} aNode The Node used to determine which portion of the
     *     screen to take a snapshot of. Note that if this is not supplied, the
     *     whole window of the currently executing window context will be used.
     * @returns {TP.test.GUIDriver} The receiver.
     */

    if (!TP.sys.isHeadless()) {
        return this.raise('TP.sig.UnsupportedOperation');
    }

    return TP.todo();
});

//  ------------------------------------------------------------------------

TP.lang.Object.defineSubtype('test.GUISequence');

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineAttribute('sequenceEntries');
TP.test.GUISequence.Inst.defineAttribute('driver');
TP.test.GUISequence.Inst.defineAttribute('$currentElement');

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('init',
function(driver) {

    /**
     * @method init
     * @summary Initialize the instance.
     * @param {TP.test.GUIDriver} driver The GUI driver which created this
     *     sequence and is being used in conjunction with it to drive the GUI.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.callNextMethod();

    this.set('$currentElement', driver.getFocusedElement());

    this.set('sequenceEntries', TP.ac());
    this.set('driver', driver);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('altKeyDown',
function(keyLocation) {

    /**
     * @method altKeyDown
     * @summary Simulate the act of pressing the 'Alt' key down.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Alt' key
     *     pressed down over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyDown('Alt', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('altKeyUp',
function(keyLocation) {

    /**
     * @method altKeyUp
     * @summary Simulate the act of releasing the 'Alt' key up.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Alt' key
     *     released up over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyUp('Alt', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('click',
function(mouseButton, mouseLocation) {

    /**
     * @method click
     * @summary Simulates the act of clicking the mouse button.
     * @description Note that this method has a notion of a 'currently focused'
     *     element, but it is important to note that this will be the element
     *     that is currently focused *when the sequence that this method belongs
     *     to is executed* (which won't happen until run() is called).
     * @param {String} [mouseButton] A mouse button constant.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} [mouseLocation]
     *     The mouse target location, given as either a target Element, an
     *     AccessPath that can be used to find the element, or as a Point that
     *     will be used with the sequence's current element.
     *     If this parameter is not supplied or null, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var point,
        button,
        target,

        eventName;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(mouseLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(mouseLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(mouseLocation);
    } else if (mouseLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.gui.Point)) {
        //  Otherwise, if a TP.gui.Point has been supplied then we set the
        //  target to TP.CURRENT, which tells the sequence performing routine to
        //  use whatever the currently focused element is at the time it is
        //  executing the sequence.
        point = mouseLocation;
        target = TP.CURRENT;
    } else if (mouseLocation === TP.FOCUSED) {
        target = TP.FOCUSED;
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(button)) {
        button = mouseButton;
    }

    if (button === TP.RIGHT) {
        eventName = 'rightclick';
    } else {
        eventName = 'click';
    }

    //  If we got a valid TP.gui.Point supplied as the mouse location, then we
    //  go ahead and supply it as the argument for this step in the sequence.
    //  Otherwise, it will have to be computed from the target in the sequence
    //  performing routine.
    if (TP.isValid(point)) {
        this.get('sequenceEntries').add(
                TP.ac(
                    eventName,
                    target,
                    {
                        pageX: point.getX(),
                        pageY: point.getY()
                    }));
    } else {
        this.get('sequenceEntries').add(
                TP.ac(eventName, target));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('constructEventFunctions',
function(aSequence) {

    var thisref,

        makeEventFunc,

        eventFuncList,

        pointerEvents,
        touchEvents,

        focusableQuery,

        clicker,

        leni,
        i,

        seqItem,
        seqTarget,
        seqFocusedTarget,
        seqEvents,
        seqArgs,

        focusable,
        currentFocusable,

        lenj,
        j,

        seqCurrentTPElem;

    thisref = this;

    makeEventFunc = function(eventTarget, eventType, eventArgs) {

        var evtTarget,
            evtType,
            evtArgs;

        evtTarget = eventTarget;
        evtType = eventType;

        if (TP.isHash(eventArgs)) {
            evtArgs = TP.copy(eventArgs);
        } else {
            evtArgs = TP.hc();
        }

        return function() {
            var modifierKeyStates,
                newEvent;

            evtArgs.atPut('type', evtType);

            if (evtType === 'keydown' ||
                evtType === 'keypress' ||
                evtType === 'keyup') {
                modifierKeyStates = evtTarget[TP.MODIFIER_KEY_STATES];
                if (TP.isValid(modifierKeyStates)) {
                    evtArgs.atPut('ctrlKey', modifierKeyStates.at('Control'));
                    evtArgs.atPut('shiftKey', modifierKeyStates.at('Shift'));
                    evtArgs.atPut('altKey', modifierKeyStates.at('Alt'));
                    evtArgs.atPut('metaKey', modifierKeyStates.at('Meta'));
                }
            }

            newEvent = TP.documentConstructEvent(
                            TP.nodeGetDocument(evtTarget), evtArgs);

            thisref.installDefaultActionHandler(evtTarget, evtType, evtArgs);

            evtTarget.dispatchEvent(newEvent);

            thisref.removeDefaultActionHandler(evtTarget, evtType);
        };
    };

    eventFuncList = TP.ac();

    pointerEvents = false;
    touchEvents = false;

    focusableQuery = TP.computeFocusableQuery();

    clicker = function(isDblClick) {
        if (pointerEvents) {
            eventFuncList.push(makeEventFunc(seqTarget, 'pointerdown'));
        }
        if (touchEvents) {
            eventFuncList.push(makeEventFunc(seqTarget, 'touchstart'));
        }

        eventFuncList.push(makeEventFunc(seqTarget, 'mousedown'));

        if (currentFocusable) {
            seqCurrentTPElem = TP.wrap(seqFocusedTarget);
            if (TP.isKindOf(seqCurrentTPElem, TP.html.input) ||
                TP.isKindOf(seqCurrentTPElem, TP.html.textarea)) {

                eventFuncList.push(
                    makeEventFunc(seqFocusedTarget, 'change'));
            }

            eventFuncList.push(makeEventFunc(seqFocusedTarget, 'blur'));
        }

        if (focusable) {
            eventFuncList.push(makeEventFunc(seqTarget, 'focus'));
            eventFuncList.push(makeEventFunc(seqTarget, 'focusin'));
        }

        if (pointerEvents) {
            eventFuncList.push(makeEventFunc(seqTarget, 'pointerup'));
        }
        if (touchEvents) {
            eventFuncList.push(makeEventFunc(seqTarget, 'touchend'));
        }

        eventFuncList.push(makeEventFunc(seqTarget, 'mouseup'));

        if (isDblClick) {
            eventFuncList.push(makeEventFunc(seqTarget, 'dblclick'));
        } else {
            eventFuncList.push(makeEventFunc(seqTarget, 'click'));
        }
    };

    leni = aSequence.getSize();
    for (i = 0; i < leni; i++) {
        seqItem = aSequence.at(i);

        seqTarget = seqItem.at('target');
        seqFocusedTarget = seqItem.at('focusedTarget');
        seqEvents = seqItem.at('events');
        seqArgs = seqItem.at('args');

        focusable =
            TP.elementMatchesCSS(seqTarget, focusableQuery);
        currentFocusable =
            TP.elementMatchesCSS(seqFocusedTarget, focusableQuery);

        switch (seqItem.at('type')) {
            case 'click':
                clicker(false);
                break;

            case 'dblclick':
                clicker(false);
                clicker(true);
                break;

            case 'rightclick':
                if (pointerEvents) {
                    eventFuncList.push(makeEventFunc(seqTarget, 'pointerdown'));
                }

                eventFuncList.push(makeEventFunc(seqTarget, 'mousedown'));

                if (pointerEvents) {
                    eventFuncList.push(makeEventFunc(seqTarget, 'pointerup'));
                }

                eventFuncList.push(makeEventFunc(seqTarget, 'mouseup'));

                eventFuncList.push(makeEventFunc(seqTarget, 'contextmenu'));

                break;

            case 'key':
                lenj = seqEvents.getSize();
                for (j = 0; j < lenj; j++) {
                    eventFuncList.push(
                            makeEventFunc(seqTarget, seqEvents.at(j), seqArgs));
                }
                break;

            case 'keydown':
            case 'keyup':
            case 'mousedown':
            case 'mouseup':
                eventFuncList.push(
                        makeEventFunc(seqTarget, seqItem.at('type'), seqArgs));
                break;

            case 'sendevent':
                eventFuncList.push(
                        makeEventFunc(seqTarget,
                                        seqItem.at('args').at('type'),
                                        seqArgs));
                break;

            default:
                break;
        }
    }

    return eventFuncList;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('ctrlKeyDown',
function(keyLocation) {

    /**
     * @method ctrlKeyDown
     * @summary Simulate the act of pressing the 'Control' key down.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Control' key
     *     pressed down over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyDown('Control', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('ctrlKeyUp',
function(keyLocation) {

    /**
     * @method ctrlKeyUp
     * @summary Simulate the act of releasing the 'Control' key up.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Control' key
     *     released up over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyUp('Control', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('doubleClick',
function(mouseButton, mouseLocation) {

    /**
     * @method doubleClick
     * @summary Simulates the act of double clicking the mouse button.
     * @description Note that this method has a notion of a 'currently focused'
     *     element, but it is important to note that this will be the element
     *     that is currently focused *when the sequence that this method belongs
     *     to is executed* (which won't happen until the 'run()' method is
     *     called).
     * @param {String} [mouseButton] A mouse button constant.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} [mouseLocation]
     *     The mouse target location, given as either a target Element, an
     *     AccessPath that can be used to find the element, or as a Point that
     *     will be used with the sequence's current element.
     *     If this parameter is not supplied or null, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var point,
        button,
        target,

        eventName;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(mouseLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(mouseLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(mouseLocation);
    } else if (mouseLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.gui.Point)) {
        //  Otherwise, if a TP.gui.Point has been supplied then we set the
        //  target to TP.CURRENT, which tells the sequence performing routine to
        //  use whatever the currently focused element is at the time it is
        //  executing the sequence.
        point = mouseLocation;
        target = TP.CURRENT;
    } else if (mouseLocation === TP.FOCUSED) {
        target = TP.FOCUSED;
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(button)) {
        button = mouseButton;
    }

    if (button === TP.RIGHT) {
        eventName = 'rightclick';
    } else {
        eventName = 'dblclick';
    }

    eventName = 'dblclick';

    //  If we got a valid TP.gui.Point supplied as the mouse location, then we
    //  go ahead and supply it as the argument for this step in the sequence.
    //  Otherwise, it will have to be computed from the target in the sequence
    //  performing routine.
    if (TP.isValid(point)) {
        this.get('sequenceEntries').add(
                TP.ac(
                    eventName,
                    target,
                    {
                        pageX: point.getX(),
                        pageY: point.getY()
                    }));
    } else {
        this.get('sequenceEntries').add(
                TP.ac(eventName, target));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('exec',
function(aFunction) {

    /**
     * @method exec
     * @summary Executes the supplied Function as a step in the sequence.
     * @param {Function} aFunction The Function to execute.
     * @returns {TP.test.GUISequence} The receiver.
     */

    if (TP.isCallable(aFunction)) {
        this.get('sequenceEntries').add(TP.ac('exec', aFunction));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('$expandSequenceEntries',
function(entries) {

    /**
     * @method $expandSequenceEntries
     * @summary Expands the supplied Array of sequence entries into a set
     *     where, if an entry's 'target' reference expands into more than one
     *     target element, an individual entry is made for each one of those
     *     targets.
     * @param {Node|TP.dom.Node|TP.path.AccessPath|Array} entries The initial
     *     set of sequence entries.
     * @returns {Node[][]|TP.dom.Node[][]} The fully expanded Array of sequence
     *     entries.
     */

    var driver,

        newEntries,

        len,
        i,

        entry,

        targets;

    driver = this.get('driver');

    newEntries = TP.ac();

    //  Loop over each entry and expand it to more than one entry if it's target
    //  resolves to more than one target.

    len = entries.getSize();
    for (i = 0; i < len; i++) {
        entry = entries.at(i);

        targets = entry.at(1);

        if (TP.isNode(targets)) {
            //  The target is already a Node
            void 0;
        } else if (TP.isKindOf(targets, TP.dom.Node)) {
            targets = TP.unwrap(targets);
        } else if (targets.isAccessPath()) {
            targets = targets.executeGet(driver.getCurrentNativeDocument());
        }

        if (TP.isArray(targets)) {
            /* eslint-disable no-loop-func */
            targets.perform(
                function(aTarget) {
                    newEntries.push(TP.ac(entry.at(0), aTarget, entry.at(2)));
                });
            /* eslint-enable no-loop-func */
        } else {
            newEntries.push(TP.ac(entry.at(0), targets, entry.at(2)));
        }
    }

    return newEntries;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('keyDown',
function(aKey, keyLocation) {

    /**
     * @method keyDown
     * @summary Simulate the act of pressing the supplied key down.
     * @param {String} aKey The key to simulate pressing down.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} keyLocation
     *     The key target location, given as either a target Element, or as an
     *     AccessPath that can be used to find the element. If this isn't
     *     supplied, the sequence's current element will be used as the target
     *     for this event. The sequence's current element will default to the
     *     element currently focused when the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var target,
        key;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(keyLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(keyLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(keyLocation);
    } else if (keyLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, String)) {
        if (keyLocation === TP.FOCUSED) {
            target = TP.FOCUSED;
        } else {
            //  Otherwise, set the target to TP.CURRENT, which tells the
            //  sequence performing routine to use whatever the sequence's
            //  current element is at the time it is executing the sequence.
            target = TP.CURRENT;
        }
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(key)) {
        key = aKey;
    }

    this.get('sequenceEntries').add(TP.ac('keydown', target, key));

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('keyUp',
function(aKey, keyLocation) {

    /**
     * @method keyUp
     * @summary Simulate the act of releasing the supplied key up.
     * @param {String} aKey The key to simulate releasing up.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} keyLocation
     *     The key target location, given as either a target Element, or as an
     *     AccessPath that can be used to find the element. If this isn't
     *     supplied, the sequence's current element will be used as the target
     *     for this event. The sequence's current element will default to the
     *     element currently focused when the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var target,
        key;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(keyLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(keyLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(keyLocation);
    } else if (keyLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, String)) {
        if (keyLocation === TP.FOCUSED) {
            target = TP.FOCUSED;
        } else {
            //  Otherwise, set the target to TP.CURRENT, which tells the
            //  sequence performing routine to use whatever the sequence's
            //  current element is at the time it is executing the sequence.
            target = TP.CURRENT;
        }
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(key)) {
        key = aKey;
    }

    this.get('sequenceEntries').add(TP.ac('keyup', target, key));

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('metaKeyDown',
function(keyLocation) {

    /**
     * @method metaKeyDown
     * @summary Simulate the act of pressing the 'Meta' key down.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Meta' key
     *     pressed down over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyDown('Meta', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('metaKeyUp',
function(keyLocation) {

    /**
     * @method metaKeyUp
     * @summary Simulate the act of releasing the 'Meta' key up.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Meta' key
     *     released up over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyUp('Meta', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('mouseDown',
function(mouseButton, mouseLocation) {

    /**
     * @method mouseDown
     * @summary Simulates the act of pressing the mouse button down.
     * @description Note that this method has a notion of a 'currently focused'
     *     element, but it is important to note that this will be the element
     *     that is currently focused *when the sequence that this method belongs
     *     to is executed* (which won't happen until the 'run()' method is
     *     called).
     * @param {String} [mouseButton] A mouse button constant.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} [mouseLocation]
     *     The mouse target location, given as either a target Element, an
     *     AccessPath that can be used to find the element, or a Point that will
     *     be used with the sequence's current element.
     *     If this parameter is not supplied or null, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var point,
        button,
        target,

        eventName;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(mouseLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(mouseLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(mouseLocation);
    } else if (mouseLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.gui.Point)) {
        //  Otherwise, if a TP.gui.Point has been supplied then we set the
        //  target to TP.CURRENT, which tells the sequence performing routine to
        //  use whatever the currently focused element is at the time it is
        //  executing the sequence.
        point = mouseLocation;
        target = TP.CURRENT;
    } else if (mouseLocation === TP.FOCUSED) {
        target = TP.FOCUSED;
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(button)) {
        button = mouseButton;
    }

    eventName = 'mousedown';

    //  If we got a valid TP.gui.Point supplied as the mouse location, then we
    //  go ahead and supply it as the argument for this step in the sequence.
    //  Otherwise, it will have to be computed from the target in the sequence
    //  performing routine.
    if (TP.isValid(point)) {
        this.get('sequenceEntries').add(
                TP.ac(
                    eventName,
                    target,
                    {
                        pageX: point.getX(),
                        pageY: point.getY()
                    }));
    } else {
        this.get('sequenceEntries').add(
                TP.ac(eventName, target));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('mouseUp',
function(mouseButton, mouseLocation) {

    /**
     * @method mouseUp
     * @summary Simulates the act of releasing the mouse button up.
     * @description Note that this method has a notion of a 'currently focused'
     *     element, but it is important to note that this will be the element
     *     that is currently focused *when the sequence that this method belongs
     *     to is executed* (which won't happen until the 'run()' method is
     *     called).
     * @param {String} [mouseButton] A mouse button constant.
     * @param {Element|TP.path.AccessPath|TP.gui.Point|Constant} [mouseLocation]
     *     The mouse target location, given as either a target Element, an
     *     AccessPath that can be used to find the element, or as a Point that
     *     will be used with the sequence's current element.
     *     If this parameter is not supplied or null, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var point,
        button,
        target,

        eventName;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(mouseLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(mouseLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(mouseLocation);
    } else if (mouseLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = mouseLocation;
    } else if (TP.isKindOf(mouseLocation, TP.gui.Point)) {
        //  Otherwise, if a TP.gui.Point has been supplied then we set the
        //  target to TP.CURRENT, which tells the sequence performing routine to
        //  use whatever the currently focused element is at the time it is
        //  executing the sequence.
        point = mouseLocation;
        target = TP.CURRENT;
    } else if (mouseLocation === TP.FOCUSED) {
        target = TP.FOCUSED;
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(button)) {
        button = mouseButton;
    }

    eventName = 'mouseup';

    //  If we got a valid TP.gui.Point supplied as the mouse location, then we
    //  go ahead and supply it as the argument for this step in the sequence.
    //  Otherwise, it will have to be computed from the target in the sequence
    //  performing routine.
    if (TP.isValid(point)) {
        this.get('sequenceEntries').add(
                TP.ac(
                    eventName,
                    target,
                    {
                        pageX: point.getX(),
                        pageY: point.getY()
                    }));
    } else {
        this.get('sequenceEntries').add(
                TP.ac(eventName, target));
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('run',
function() {

    /**
     * @method run
     * @summary Executes the sequence.
     * @description Note that the steps that comprise the sequence are all
     *     executed in a 'chained' manner, so that if there are asynchronous
     *     delays in certain actions, they will all be executed in the proper
     *     order. Also note that this method generates a Promise, which is then
     *     chained to the driver's Promise supplier's internal Promise, so that
     *     if there are other asynchronous actions that are using Promises from
     *     the receiver's driver's Promise supplier, this method will respect
     *     that.
     * @returns {Promise} A Promise which completes when the event sequence is
     *     complete.
     */

    var thisref,
        promise;

    thisref = this;

    //  A Promise that will execute set up a 'work' callback function that
    //  executes each entry in the sequence and then resolves the Promise after
    //  executing the last entry.
    promise = Promise.construct(
                function(resolver, rejector) {

                    var currentElement,
                        sequenceEntries,

                        driver,

                        count,
                        workFunc;

                    currentElement = thisref.get('$currentElement');

                    sequenceEntries = thisref.get('sequenceEntries');
                    driver = thisref.get('driver');

                    //  'Expand' any event targets in the sequence entries
                    sequenceEntries = thisref.$expandSequenceEntries(
                                                        sequenceEntries);

                    //  Set up the work function that will process a single
                    //  entry and then supply a callback that will call back the
                    //  work function, but only after a delay to give the GUI a
                    //  chance to refresh.
                    count = 0;
                    workFunc = function() {

                        var seqEntry,

                            focusedElement,
                            defaultElement,

                            workCallback;

                        //  If the count equals the number of entries, then
                        //  we're done here and we can resolve the Promise.
                        if (count === sequenceEntries.getSize()) {

                            sequenceEntries = null;

                            return resolver();
                        }

                        seqEntry = sequenceEntries.at(count);
                        count++;

                        //  If it's an 'exec', then we're just executing a
                        //  Function. Execute it and then make sure to call the
                        //  resolver.
                        if (seqEntry.at(0) === 'exec') {
                            seqEntry.at(1)();

                            return workFunc();
                        } else {

                            focusedElement = driver.getFocusedElement();

                            //  If the sequence doesn't have a current element,
                            //  then try to get the currently focused element.
                            if (!TP.isElement(currentElement)) {

                                defaultElement = focusedElement;

                                //  If we can't determine a focused element,
                                //  call the error callback and exit.
                                if (!TP.isElement(defaultElement)) {

                                    sequenceEntries = null;

                                    return rejector(
                                    'No default Element for the GUI Driver.');
                                }
                            } else {
                                defaultElement = currentElement;
                            }

                            //  We fork the work function here to give the GUI
                            //  a chance to refresh before we manipulate it.
                            workCallback =
                                function() {
                                    workFunc.queueAfterNextRepaint(
                                        TP.nodeGetWindow(defaultElement));
                                };

                            //  Execute the individual sequence step entry.
                            thisref.$performGUISequenceStep(
                                seqEntry.at(1),
                                seqEntry.at(0),
                                seqEntry.at(2),
                                workCallback,
                                defaultElement,
                                focusedElement);
                        }
                    };

                    workFunc();
                });

    if (TP.isArray(TP.signal.$deferreds)) {
        promise = promise.then(
                    function() {
                        return TP.signal.resolveDeferreds();
                    });
    }

    return promise;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('$performGUISequenceStep',
function(target, type, args, callback, defaultElement, focusedElement) {

    /**
     * @method $performGUISequenceStep
     * @summary Actually performs the sequence step, firing an event at the
     *     supplied target element.
     * @param {Element} target The element to target the event at.
     * @param {String} type The type of event to generate.
     * @param {Object} args The configuration argument used to build the event.
     * @param {Function} callback The callback function to execute when the
     *     event dispatch is complete.
     * @param {Element} defaultElement The default Element that events should be
     *     directed to, if the target is not defined or is explicitly set to
     *     TP.CURRENT.
     * @param {Element} focusedElement The currently focused Element. This may
     *     be the same as the focused element, but not if the sequence contains
     *     a whole number of steps that all need to be directed to the same
     *     element.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var makeKeyRecord,

        finalTarget,

        eventFuncList,

        leni,
        i,

        keyMatcher,

        keyArgs,

        results,
        keyResults,

        lenj,
        j,

        val,

        chars,

        lenk,
        k,

        resultRecord,
        resultKey,

        resultEvents;

    makeKeyRecord = function(keyStr) {
        var record;

        switch (keyStr) {
            case '[Shift]':
                record = TP.hc('events', TP.ac('keydown'),
                                'eventKey', 'Shift');
                break;
            case '[Shift-Up]':
                record = TP.hc('events', TP.ac('keyup'),
                                'eventKey', 'Shift');
                break;
            case '[Control]':
                record = TP.hc('events', TP.ac('keydown'),
                                'eventKey', 'Control');
                break;
            case '[Control-Up]':
                record = TP.hc('events', TP.ac('keyup'),
                                'eventKey', 'Control');
                break;
            case '[Alt]':
                record = TP.hc('events', TP.ac('keydown'),
                                'eventKey', 'Alt');
                break;
            case '[Alt-Up]':
                record = TP.hc('events', TP.ac('keyup'),
                                'eventKey', 'Alt');
                break;
            case '[Meta]':
                record = TP.hc('events', TP.ac('keydown'),
                                'eventKey', 'Meta');
                break;
            case '[Meta-Up]':
                record = TP.hc('events', TP.ac('keyup'),
                                'eventKey', 'Meta');
                break;
            case '[CapsLock]':
                record = TP.hc('events', TP.ac('keydown'),
                                'eventKey', 'CapsLock');
                break;
            case '[CapsLock-Up]':
                record = TP.hc('events', TP.ac('keyup'),
                                'eventKey', 'CapsLock');
                break;
            case '[Home]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Home');
                break;
            case '[End]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'End');
                break;
            case '[PageUp]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'PageUp');
                break;
            case '[PageDown]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'PageDown');
                break;
            case '[ArrowLeft]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'ArrowLeft');
                break;
            case '[ArrowRight]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'ArrowRight');
                break;
            case '[ArrowUp]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'ArrowUp');
                break;
            case '[ArrowDown]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'ArrowDown');
                break;
            case '[Backspace]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Backspace');
                break;
            case '[Insert]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Insert');
                break;
            case '[Delete]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Delete');
                break;
            case '[Enter]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Enter');
                break;
            case '[Escape]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Escape');
                break;
            case '[Tab]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'Tab');
                break;
            case '[F1]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F1');
                break;
            case '[F2]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F2');
                break;
            case '[F3]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F3');
                break;
            case '[F4]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F4');
                break;
            case '[F5]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F5');
                break;
            case '[F6]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F6');
                break;
            case '[F7]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F7');
                break;
            case '[F8]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F8');
                break;
            case '[F9]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F9');
                break;
            case '[F10]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F10');
                break;
            case '[F11]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F11');
                break;
            case '[F12]':
                record = TP.hc('events', TP.ac('keydown', 'keyup'),
                                'eventKey', 'F12');
                break;
            default:
                //  Alphanumeric keys
                record = TP.hc('events', TP.ac('keydown', 'keypress', 'keyup'),
                                'eventKey', keyStr);

                break;
        }

        return record;
    };

    if (target === TP.CURRENT) {
        finalTarget = defaultElement;
    } else if (target === TP.FOCUSED) {
        finalTarget = focusedElement;
    } else if (TP.isElement(target)) {
        finalTarget = target;
    } else if (target.isAccessPath()) {
        //  TODO: Implement access path
    } else {
        return this.raise('TP.sig.InvalidTarget',
                            'Bad target in GUI driver:: ' + TP.str(target));
    }

    eventFuncList = TP.ac();

    //  The three parameters are:
    //      1. The type of operation, used in the switch below.
    //      2. The target of the operation. This could be an Element specified
    //      by the user, an Element derived from an access path given by the
    //      user (multiple Elements will have been expanded into multiple
    //      entries or the currently focused Element as supplied to this method.
    //      3. Operation-specific arguments.

    /* eslint-disable no-fallthrough */
    switch (type) {
        case 'click':
            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', 'click')
                            ));
            break;

        case 'dblclick':
            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', 'dblclick')
                            ));
            break;

        case 'rightclick':

            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', 'rightclick')
                            ));

            break;

        case 'key':
        case 'keys':
            keyMatcher = /(\[[a-zA-Z0-9-]+?\]|[^[]?[^[]])/;
            if (!TP.isArray(args)) {
                keyArgs = TP.ac(args);
            } else {
                keyArgs = args;
            }

            leni = keyArgs.getSize();
            for (i = 0; i < leni; i++) {
                results = keyArgs.at(i).split(keyMatcher);

                keyResults = TP.ac();

                lenj = results.getSize();
                for (j = 0; j < lenj; j++) {
                    val = TP.trim(results.at(j));
                    if (/^\[.+\]$/.test(val)) {
                        keyResults.push(makeKeyRecord(val));
                    } else {
                        chars = val.split('');
                        lenk = chars.getSize();
                        for (k = 0; k < lenk; k++) {
                            keyResults.push(makeKeyRecord(chars.at(k)));
                        }
                    }
                }
            }

            leni = keyResults.getSize();
            for (i = 0; i < leni; i++) {
                resultRecord = keyResults.at(i);

                resultKey = resultRecord.at('eventKey');
                resultEvents = resultRecord.at('events');

                eventFuncList = eventFuncList.concat(
                                this.constructEventFunctions(
                                    TP.ac(
                                        TP.hc(
                                            'target', finalTarget,
                                            'focusedTarget', focusedElement,
                                            'type', 'key',
                                            'events', resultEvents,
                                            'args', TP.hc('key', resultKey))
                                    )));
            }

            break;

        case 'keydown':
        case 'keyup':

            resultRecord = makeKeyRecord(args);
            resultKey = resultRecord.at('eventKey');

            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', type,
                                        'args', TP.hc('key', resultKey))
                            ));
            break;

        case 'mousedown':
        case 'mouseup':
            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', type)
                            ));
            break;

        case 'sendevent':
            eventFuncList = this.constructEventFunctions(
                            TP.ac(
                                TP.hc('target', finalTarget,
                                        'focusedTarget', focusedElement,
                                        'type', 'sendevent',
                                        'args', TP.hc('type', args.at('type')))
                            ));
            break;

        default:
            break;
    }
    /* eslint-enable no-fallthrough */

    leni = eventFuncList.getSize();
    for (i = 0; i < leni; i++) {
        setTimeout(eventFuncList.at(i), 0);
    }

    setTimeout(callback, 0);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('installDefaultActionHandler',
function(anElement, eventType, eventArgs) {

    var handlerName,

        defaultActionHandlers,

        tpElem;

    handlerName = 'performDefaultActionFor' + TP.makeStartUpper(eventType);

    defaultActionHandlers = anElement[TP.DEFAULT_ACTION_HANDLERS];
    if (TP.notValid(defaultActionHandlers)) {
        defaultActionHandlers = TP.hc();
        anElement[TP.DEFAULT_ACTION_HANDLERS] = defaultActionHandlers;
    }

    tpElem = TP.wrap(anElement);

    if (TP.canInvoke(tpElem, handlerName) &&
        !defaultActionHandlers.hasKey(eventType)) {

        (function() {
            var handler;

            handler = async function(evt) {
                await tpElem[handlerName](evt);
            };

            TP.boot.$$addNativeHandler(anElement, eventType, handler, false);
            defaultActionHandlers.atPut(eventType, handler);
        }());
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('removeDefaultActionHandler',
function(anElement, eventType) {

    var defaultActionHandlers,

        handler;

    defaultActionHandlers = anElement[TP.DEFAULT_ACTION_HANDLERS];
    if (TP.notValid(defaultActionHandlers)) {
        return this;
    }

    handler = defaultActionHandlers.at(eventType);
    if (TP.isValid(handler)) {
        TP.boot.$$removeNativeHandler(anElement, eventType, handler, false);
    }

    defaultActionHandlers.removeKey(eventType);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('sendEvent',
function(eventInfo, eventLocation) {

    /**
     * @method sendEvent
     * @summary Sends the event as described in the supplied eventInfo. See the
     *     TP.documentConstructEvent() call for more information on the fields
     *     recognized ere.
     * @param {TP.core.Hash|Object|String} eventInfo The event information used
     *     to create the event to send or a simple event name which will be
     *     used to create an eventInfo object via TP.hc('type', eventInfo).
     * @param {Element|TP.path.AccessPath} eventLocation The target element or
     *     an access path to the target element that should be the target of the
     *     event. If this isn't supplied, the sequence's current element will be
     *     used as the target for this event. The sequence's current element
     *     will default to the element currently focused when the sequence
     *     begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var target,
        info;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(eventLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(eventLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = eventLocation;
    } else if (TP.isKindOf(eventLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(eventLocation);
    } else if (eventLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = eventLocation;
    } else if (eventLocation === TP.FOCUSED) {
        target = TP.FOCUSED;
    } else {
        //  TODO: Raise an exception
        return this;
    }

    info = TP.isString(eventInfo) ? TP.hc('type', eventInfo) : eventInfo;

    this.get('sequenceEntries').add(TP.ac('sendevent', target, info));

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('sendKeys',
function(aString, keyLocation) {

    /**
     * @method sendKeys
     * @summary Simulate the act of pressing and releasing the sequence of keys
     *     necessary to reproduce the effect of the user typing what is
     *     specified in the String.
     * @description Note that it is possible to supply W3C-defined control
     *     keys surrounded with brackets ('[' and ']') in this String for ease
     *     of sending these control characters to the GUI:
     *          'ABC[Left][Backspace]D[Right]E'
     *
     *     This will produce 'ADCE' in the GUI.
     *
     *     Also note that appending '-up' to some of these keys, such as
     *     'Shift', 'Control', 'Alt' and 'Meta', will simulate the act of
     *     releasing the key:
     *          '[Alt]a[Alt-up]'
     *     Note that all of the possible keys that can be used here can be found
     *     in a TIBET keyboard map as 'key="..."' entries.
     * @param {String} String The String used to determine which keys will be
     *     pressed and released.
     * @param {Element|TP.path.AccessPath} [keyLocation] The target element or
     *     an access path to the target element that should have the keys
     *     pressed down and released up over. If this isn't supplied, the
     *     currently focused element in the receiver's owning driver's window
     *     context will be used as the target for this event.
     * @returns {TP.test.GUISequence} The receiver.
     */

    var target,
        keys;

    //  If there was no valid location supplied, then we just set the target to
    //  TP.CURRENT, which means to use the currently focused element whenever
    //  the sequence is executed.
    if (TP.notValid(keyLocation)) {
        target = TP.CURRENT;
    } else if (TP.isElement(keyLocation)) {
        //  Otherwise, if an Element has been supplied, use that.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, TP.dom.ElementNode)) {
        //  Otherwise, if a TP.dom.ElementNode has been supplied, unwrap and use
        //  that.
        target = TP.unwrap(keyLocation);
    } else if (keyLocation.isAccessPath()) {
        //  Otherwise, if a TP.path.AccessPath has been supplied, that will be
        //  used to determine the target element.
        target = keyLocation;
    } else if (TP.isKindOf(keyLocation, String)) {
        if (keyLocation === TP.FOCUSED) {
            target = TP.FOCUSED;
        } else {
            //  Otherwise, if a String has been supplied then we set the target
            //  to TP.CURRENT, which tells the sequence performing routine to
            //  use whatever the currently focused element is at the time it is
            //  executing the sequence.
            keys = keyLocation;
            target = TP.CURRENT;
        }
    } else {
        //  TODO: Raise an exception
        return this;
    }

    if (TP.notValid(keys)) {
        keys = aString;
    }

    this.get('sequenceEntries').add(TP.ac('keys', target, keys));

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('shiftKeyDown',
function(keyLocation) {

    /**
     * @method shiftKeyDown
     * @summary Simulate the act of pressing the 'Shift' key down.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Shift' key
     *     pressed down over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyDown('Shift', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------

TP.test.GUISequence.Inst.defineMethod('shiftKeyUp',
function(keyLocation) {

    /**
     * @method shiftKeyUp
     * @summary Simulate the act of releasing the 'Shift' key up.
     * @param {Element|TP.path.AccessPath} keyLocation The target element or an
     *     access path to the target element that should have the 'Shift' key
     *     released up over. If this isn't supplied, the sequence's current
     *     element will be used as the target for this event. The sequence's
     *     current element will default to the element currently focused when
     *     the sequence begins.
     * @returns {TP.test.GUISequence} The receiver.
     */

    this.keyUp('Shift', keyLocation);

    return this;
});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
