//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 * @overview The 'tibet watch' command. Watches one or more directories and
 *     runs a tibet command/task for any file changes in those directories.
 */
//  ========================================================================

/* eslint indent:0 */

(function() {

'use strict';

var CLI,
    chokidar,
    Cmd,
    Promise,
    path;

CLI = require('./_cli');
chokidar = require('chokidar');
path = require('path');
Promise = require('bluebird');

//  ---
//  Type Construction
//  ---

Cmd = function() { /* init */ };
Cmd.Parent = require('./_cmd');
Cmd.prototype = new Cmd.Parent();


//  ---
//  Type Attributes
//  ---

/**
 * The command execution context.
 * @type {Cmd.CONTEXTS}
 */
Cmd.CONTEXT = CLI.CONTEXTS.INSIDE;

/**
 * The command name for this type.
 * @type {string}
 */
Cmd.NAME = 'watch';

//  ---
//  Instance Attributes
//  ---

/* eslint-disable quote-props */
Cmd.prototype.PARSE_OPTIONS = CLI.blend({
    string: ['include', 'exclude'],
    default: {}
},
Cmd.Parent.prototype.PARSE_OPTIONS);
/* eslint-enable quote-props */
/**
 * The command usage string.
 * @type {string}
 */
Cmd.prototype.USAGE = 'tibet watch <path> <task> [--include] [--exclude]';


//  ---
//  Instance Methods
//  ---

/**
 * Perform the actual command processing logic.
 * @returns {Number|Promise} The return code produced by running the command (a
 *     non-zero indicates an Error) or a Promise that resolves when the command
 *     finishes.
 */
Cmd.prototype.execute = function() {
    var root,
        task,
        thisref,
        pkg,
        base,
        exclude,
        escaper,
        pattern,
        watcher,
        promise;

    pkg = CLI.getPackage();
    base = CLI.expandPath('~');

    root = this.getArgv()[1];
    if (CLI.isEmpty(root)) {
        throw new Error('No watch root directory.');
    }
    root = root.trim();
    root = root === '.' ? process.cwd() : root;

    root = root.split('|').map((item) => {
        return CLI.expandPath(item);
    });

    root = root.map((item) => {
        if (!CLI.sh.test('-d', item)) {
            CLI.warn(`Watch root ${item} not found.`);
            return '';
        }
        return path.resolve(pkg.expandPath(item));
    }).filter((item) => {
        return item;
    });

    task = this.getArgv()[2];
    if (CLI.isEmpty(task)) {
        throw new Error('No watcher task to run.');
    }
    task = task.trim();

    exclude = this.options.exclude;

    //  Helper function for escaping regex metacharacters. NOTE
    //  that we need to take "ignore format" things like path/*
    //  and make it path/.* or the regex will fail. Also note we special
    //  case ~ to allow virtual path matches.
    escaper = function(str) {
        return str.replace(
            /\*/g, '.*').replace(
            /\./g, '\\.').replace(
            /\~/g, '\\~').replace(
            /\//g, '\\/');
    };

    if (exclude) {
        pattern = exclude.reduce(function(str, item) {
            var fullpath;

            fullpath = CLI.expandPath(item);
            return str ?
                str + '|' + escaper(fullpath) :
                escaper(fullpath);
        }, '');

        pattern += '|\\.git|\\.svn|node_modules|[\\/\\\\]\\..';

        try {
            pattern = new RegExp(pattern);
        } catch (e) {
            return CLI.error('Error creating RegExp: ' +
                e.message);
        }
    } else {
        pattern = /\.git|\.svn|node_modules|[\/\\]\../;
    }

    promise = Promise.pending();

    CLI.debug(`${task} watching ${JSON.stringify(root)}`);
    CLI.debug(`excluding: ${pattern}`);

    watcher = chokidar.watch(root, {
        ignored: pattern,
        cwd: base,
        alwaysStat: true,
        interval: 100,
        ignoreInitial: true,
        ignorePermissionErrors: true,
        persistent: true
    });

    thisref = this;

    //  The primary change handler function.
    watcher.on('change', function(file, stats) {
        var tibetpath,
            result,
            cmdline;

        CLI.debug(stats);

        //  NOTE closure'd root var here
        tibetpath = CLI.getVirtualPath(CLI.joinPaths(base, file));

        cmdline = `tibet ${task} '${tibetpath}'`;
        CLI.info(`${cmdline}`);

        //  Perform any extension-specific processing for the file.
        try {
            result = thisref.shexec(cmdline, false);
            CLI.info(result);
        } catch (e) {
            CLI.error(`Error running ${cmdline}:
${e.message}`);
            promise.reject(e);
        }
    });

    return promise;
};


module.exports = Cmd;

}());
