//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ------------------------------------------------------------------------

/**
 * @type {TP.ux.Element}
 * @summary Common mixin trait for CodeRats UX controls. CodeRats apps rely
 *     on newly designed templating and data-binding subsystems for TIBET
 *     which are the primary focus of this mixin.
 */

//  ------------------------------------------------------------------------

TP.dom.NonNativeUIElementNode.defineSubtype('TP.ux.Element');

/*
TP.ux.Element.Type.resolveTrait(
        'tagExpand',
        TP.dom.UIElementNode);
*/

TP.ux.Element.Inst.resolveTraits(
        TP.ac('$setAttribute', 'removeAttribute', 'select', 'signal'),
        TP.dom.UIElementNode);

//  ------------------------------------------------------------------------
//  Type Attributes
//  ------------------------------------------------------------------------

//  TODO:   future updates should remove the requirement for these.
TP.ux.Element.defineAttribute('styleURI', TP.NO_RESULT);
TP.ux.Element.Type.defineAttribute('themeURI', TP.NO_RESULT);

//  Templated subtypes (and optimized/factored tags) rely on an expansion
//  transform function during tagExpand processing. The tagCompile method
//  is typically responsible for creating this.
TP.ux.Element.Type.defineAttribute('expansionTransform');

//  ------------------------------------------------------------------------
//  Type Methods
//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('tagAttachBinds',
function(aRequest) {

    /**
     * @method tagAttachBinds
     * @summary Awakens bind: namespace event handlers for the element in
     *     aRequest. This method is invoked in response to mutation events
     *     when the element(s) in question are added to a visible DOM.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,

        query,
        boundElements;

    //  Make sure that we have a node to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    //  Note that the :scope here isn't "bind:scope" it's query scope :).
    query = ':scope *[*|io],' +
            ':scope *[*|in]';

    boundElements = TP.ac(elem.querySelectorAll(query));

    //  Additionally, if the supplied Element itself matches that query then
    //  unshift it onto the result.
    query = '*[*|io], *[*|in]';
    if (elem.matches(query)) {
        boundElements.unshift(elem);
    }

    //  Each element is wrapped in its own instance and observes happen there.
    boundElements.forEach(
        function(anElem) {
            var tpElem;

            tpElem = TP.wrap(anElem);
            if (!TP.canInvoke(tpElem, 'observeBindings')) {
                return;
            }
            tpElem.observeBindings();
        });

    return;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('tagAttachComplete',
function(aRequest) {

    /**
     * @method tagAttachComplete
     * @summary Sets up runtime machinery for the element in aRequest.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        tpElem;

    //  this makes sure we maintain parent processing
    this.callNextMethod();

    //  Make sure that we have a node to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    tpElem = TP.wrap(elem);
    tpElem.setup();

    return;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('tagDetachBinds',
function(aRequest) {

    /**
     * @method tagDetachBinds
     * @summary Disconnects bind: namespace event handlers for the element in
     *     aRequest. This method is invoked in response to mutation events
     *     when the element(s) in question are detached from a visible DOM.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        tpElem;

    //  Make sure that we have a node to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    tpElem = TP.wrap(elem);
    tpElem.ignoreBindings();

    return;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('tagDetachComplete',
function(aRequest) {

    /**
     * @method tagDetachComplete
     * @summary Tears down runtime machinery for the element in aRequest.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,
        tpElem;

    //  this makes sure we maintain parent processing
    this.callNextMethod();

    //  Make sure that we have a node to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    tpElem = TP.wrap(elem);
    tpElem.teardown();

    return;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('tagExpand',
async function(aRequest) {

    /**
     * @method tagExpand
     * @summary The top-level "macro expansion" method for UI tags. NOTE
     *     you normally avoid overriding this method and should leverage
     *     getExpandedNode or getExpandedSource methods (depending on
     *     circumstances) to provide this method with the final node.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {Node} The new node or fragment result from expansion.
     */

    var elem,
        tpElem,
        result,
        action,
        newNode,
        newFragment,
        isFragment,
        retVal;

    //  Make sure that we have a node to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    tpElem = TP.wrap(elem);

    //  Do the actual tag expansion via getExpandedNode.
    result = await this.getExpandedNode(tpElem, aRequest);

    if (TP.isArray(result)) {
        newNode = result.first();
        action = result.last();
    } else {
        newNode = result;
    }

    if (TP.notValid(newNode)) {
        return this.raise('TP.sig.InvalidNode');
    }

    if (TP.notValid(action)) {
        switch (newNode.nodeType) {
            case 1: //  element
                if (newNode === elem) {
                    //  identical element? assumption is there's no additional
                    //  processing necessary below the returned node.
                    action = TP.NONE;
                } else if (newNode.nodeName === elem.nodeName &&
                        TP.name(newNode) === TP.name(elem)) {
                    //  same tag (and tibet:tag?)
                    //  this happens with default injection of alert link but
                    //  could happen for a lot of other reasons. in any case the
                    //  presumption is the new node needs to replace the old
                    action = TP.REPLACE;
                } else {
                    //  different tag? assume replacement.
                    action = TP.REPLACE;
                }
                break;
            case 2:   //  attribute
                action = TP.NONE;
                break;
            case 3:   //  text node
                action = TP.UPDATE;
                break;
            case 4:   //  cdata node
                action = TP.UPDATE;
                break;
            case 5:   //  entity reference node
                action = TP.UPDATE;
                break;
            case 6:   //  entity node
                action = TP.UPDATE;
                break;
            case 7:   //  processing instruction node
                action = TP.UPDATE;
                break;
            case 8:   //  comment node
                action = TP.UPDATE;
                break;
            case 9:   //  document node
                newFragment = TP.nodeListAsFragment(newNode.children);
                newNode = newFragment;
                action = TP.UPDATE;
                break;
            case 10:  //  doctype node
                action = TP.NONE;
                break;
            case 11:  //  document fragment node
                action = TP.UPDATE;
                break;
            case 12:  //  notation node
                action = TP.UPDATE;
                break;
            default:
                action = TP.NONE;
                break;
        }
    }

    isFragment = TP.isFragment(newNode);

    //  Note we return the *result* of this method due to node importing, etc.
    switch (action) {
        case TP.NONE:
            TP.warn(`Taking no action for expansion of ${TP.name(elem)}.`);

            //  TODO:   better retVal here.
            //  NOTE: returning an array is a hack since the outer tag processor
            //  doesn't currently handle them, only nodes/fragments.
            return [];
        case TP.UPDATE:
            //  Update the element's content, but do *NOT* replace it.
            if (isFragment) {
                //  NB: nodeSetContent will return the first child that was
                //  appended, but that's not sufficient for a fragment, we
                //  need to return the full set of new nodes to process.
                await TP.nodeSetContent(elem, newNode);
                newNode = elem;
                retVal = TP.nodeGetChildNodes(newNode);
            } else {
                //  NB: if what comes back isn't an element then we can't really
                //  use it as the retVal.
                newNode = await TP.nodeSetContent(elem, newNode);
                retVal = TP.isCollectionNode(newNode) ? newNode : elem;
            }
            break;
        case TP.REPLACE:
            //  Fallthrough
        default:
            //  Replace the whole element.
            newNode = await TP.elementReplaceWith(elem, newNode);
            retVal = newNode;

            //  If we replaced the element we need to ensure things like
            //  TP.WRAPPER etc. are migrated to retain continuity.
            this.mergeExpansionAttributes(elem, newNode);
            this.mergeExpansionProperties(elem, newNode);

            break;
    }

    //  retVal is the newNode or an array containing all newly inserted nodes.
    //  We need to update any elements in that result set so they retain the
    //  processing information needed to produce the rest of the tree
    if (TP.isArray(retVal)) {
        const thisref = this;
        retVal.forEach((node) => {
            if (TP.isElement(node)) {
                newNode[TP.PROCESSOR_HINT] =
                    thisref.getExpansionTraversal(newNode, aRequest);
                TP.elementSetGenerator(newNode);
            }
        });
    } else if (TP.isElement(newNode)) {
        newNode[TP.PROCESSOR_HINT] =
            this.getExpansionTraversal(newNode, aRequest);
        TP.elementSetGenerator(newNode);
    }

    //  Return the node or node list representing the expansion.
    return retVal;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpandedErrorSource',
function(tpElement, aRequest) {

    /**
     * @method getExpandedErrorSource
     * @summary Returns a string to be produced when the receiver was unable
     *     to produce a viable expanded source string. The string returned here
     *     will be injected into the DOM as a replacement for that content.
     *     The default is a markup comment showing the tag and an error msg.
     * @param {TP.dom.UIElementNode} tpElement The wrapped instance of the
     *     receiver being processed.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {String} The content to provide for a placeholder DOM node.
     */

    var str;

    str = tpElement.getElementSource();

    return `<!-- ${str} failed expansion -->`;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpandedNode',
async function(tpElement, aRequest) {

    /**
     * @method getExpandedNode
     * @summary Creates and returns a node, fragment, or pair representing the
     *     expanded form of the tpElement parameter provided. This is the core
     *     method for producing an expanded node from a source string. Note that
     *     you normally override getExpandedSource, not this method.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {Node|Array} The new node or ordered pair result from expansion.
     *     When using an ordered pair return the new string and a processing
     *     hint (TP.UPDATE, TP.REPLACE).
     */

    var result,

        str,
        action,

        newNode;

    //  Both computed and templated tags ultimately rely on getExpandedSource to
    //  produce a string this method will process into a result or result set.
    try {
        result = await this.getExpandedSource(tpElement, aRequest);
    } catch (e) {
        TP.ifError() ? TP.error(
            TP.ec(e, `Error in ${this.getName()}.Type.getExpandedSource`)) : 0;
    }

    //  Results can always be a string and action, string, or occasionally an
    //  unusable result (due to an error in the getExpandedSource method).
    if (TP.isPair(result)) {
        str = result.first();
        action = result.last();
    } else if (TP.isString(result)) {
        str = result;
    } else {
        //  If we didn't get viable output try getting an error string
        str = this.getExpandedErrorSource(tpElement, aRequest);
    }

    //  This will create text node for simple text or elements when possible.
    newNode = TP.xhtmlnode(str);
    if (TP.notValid(newNode)) {
        TP.ifError() ? TP.error(
            `Error in ${this.getName()}.Type.getExpandedNode`) : 0;
    }

    if (TP.isValid(action)) {
        return TP.ac(newNode, action);
    }

    return newNode;
});

//  ------------------------------------------------------------------------

/* eslint-disable require-await */
TP.ux.Element.Type.defineMethod('getExpandedSource',
async function(tpElement, aRequest) {

    /**
     * @method getExpandedSource
     * @summary Creates and returns a string or pair representing the string
     *     expansion of the tpElement parameter provided. This is the core
     *     method most subtypes will override to do their expansion work.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {String|Array} The string or ordered pair result from expansion.
     *     When using an ordered pair return the new string and a processing
     *     hint (TP.UPDATE, TP.REPLACE).
     */

    var name,
        msg,
        link,
        tag;

    name = tpElement.getCanonicalName();

    msg = `'Edit the ${tpElement.getTypeName()}.Type getExpandedSource or getExpandedNode method to complete this UI.'`;

    link = '<a xmlns="' + TP.w3.Xmlns.XHTML + '" href="#" ' +
            'onclick="alert(' + msg + ')">' +
            '&lt;' + name + '/&gt;' +
            '</a>';

    tag = '<' + name + ' xmlns="' + tpElement.getNSURI() + '">' +
        link +
        '</' + name + '>';

    return tag;
});
/* eslint-enable require-await */

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpansionAttributeNames',
function() {

    /**
     * @method getExpansionAttributeNames
     * @summary Returns a list of attribute names that should be copied to any
     *     replacement nodes to retain authoring continuity. The defaults are
     *     name and class.
     * @returns {Array} The array of attribute names to retain/copy.
     */

    //  TODO:   what about tibet:ctrl etc?
    return TP.ac('name', 'class');
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpansionDictionary',
function(tpElement, aRequest) {

    /**
     * @method getExpansionDictionary
     * @summary Returns a dictionary of common variables used in templated
     *     transforms such as $TAG, $TARGET, etc.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {TP.core.Hash} A dictionary of templating variables.
     */

    var dict;

    //  NOTE we create a new dictionary with each call. If multiple async
    //  compilations were taking place we don't want mixing from a cache.
    dict = TP.hc(
        'TP', TP,
        'APP', APP,
        '$APP', TP.sys.getApplication(),

        '$TYPE', this,
        '$TAG', tpElement,

        '$REQUEST', aRequest,

        '$SOURCE', tpElement,
        '$TARGET', tpElement.getDocument(),

        '$SELECTION', tpElement.getCurrentSelectionValue.bind(tpElement),
        '$FOCUS', tpElement.getCurrentFocusedValue.bind(tpElement)
    );

    return dict;

});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpansionTransform',
function(tpElement, aRequest) {

    /**
     * @method getExpansionTransform
     * @summary Returns the receiver's expansion transform function, if any.
     *     This function is typically created during tagCompile and invoked
     *     during tagExpand processing.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {Function} The tag expansion transform function.
     */

    return this.$get('expansionTransform');
});


//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('getExpansionTraversal',
function(newNode, aRequest) {

    /**
     * @method getExpansionTraversal
     * @summary Returns the receiver's transform traversal constant. This value
     *     tells the tag processing system how to proceed, to descend, stop,
     *     or take some other approach with the newNode being returned.
     * @param {Node} newNode The node produced by the tagExpand machinery.
     *     This is the node which will ultimately be marked using the
     *     returned traversal constant.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {Number} A TP traversal constant such as TP.DESCEND.
     */

    return TP.DESCEND;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Type.defineMethod('isResponderFor',
function(aSignal, aNode) {

    /**
     * @method isResponderFor
     * @summary Returns true if the type in question should be considered a
     *     responder for the specific node/signal pair provided.
     * @param {TP.sig.Signal} aSignal The signal that responders are being
     *     computed for.
     * @param {Node} aNode The node to check which may have further data as to
     *     whether this type should be considered to be a responder.
     * @returns {Boolean} True when the receiver should respond to aSignal.
     */

    return true;
});

//  ------------------------------------------------------------------------
//  Instance Attributes
//  ------------------------------------------------------------------------

//  The list of bindings for which an observeBindings operation was done. This
//  list is populated by observeBindings and used/emptied by ignoreBindings.
TP.ux.Element.Inst.defineAttribute('$bindings');

//  Cached versions of the inbound and outbound bind entry objects. This helps
//  speed up bind operations by avoiding recomputation/parsing of bind: attrs.
TP.ux.Element.Inst.defineAttribute('$inboundBindEntries');
TP.ux.Element.Inst.defineAttribute('$outboundBindEntries');

//  UX elements are MVC-oriented so they consistently have a 'value' slot.
TP.ux.Element.Inst.defineAttribute('value');

//  ------------------------------------------------------------------------
//  Instance Methods
//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineHandler('FacetChange',
function(aSignal) {

    /**
     * @method handleFacetChange
     * @summary Responds to notifications that a facet of a change data source
     *     has changed value. Note that observations of FacetChange are setup
     *     and torn down by the observe/ignore bindings machinery.
     * @param {TP.sig.Signal} aSignal The FacetChange signal being handled.
     * @returns {TP.ux.Element} The receiver.
     */

    var bindings,
        aspect,
        checkStr,
        origin,
        entries,
        entry,
        funcName,
        value,
        current;

    //  Verify we have active bindings in place. If not we can ignore signal.
    //  NOTE that this check implies any other Change signal observations (for
    //  non-binds) must be handled via a more specific change handler.
    bindings = this.get('$bindings');
    if (TP.isEmpty(bindings)) {
        return this;
    }

    //  The signal will report which facet changed.
    aspect = aSignal.get('aspect');
    if (TP.isEmpty(aspect)) {
        return this;
    }

    //  Find any bindings referencing the path/fragment in question.
    origin = TP.gid(aSignal.getOrigin());
    entries = bindings.at(origin);
    if (TP.isEmpty(entries)) {
        return this;
    }

    checkStr = aspect.toUpperCase();

    //  Scan for at least one entry for the specific aspect...
    entry = entries.find((item) => {
        return item.first().toUpperCase() === checkStr;
    });

    //  If we didn't find the simple one look for multiply-dotted path...
    if (TP.notValid(entry)) {
        entry = entries.find((item) => {
            let expr;
            //  item.last is the kvPair, whose last is a hash containing the
            //  bind info itself (including the fullExpr
            expr = item.last().last().at('fullExpr');
            expr = expr.split('.').last().toUpperCase();

            return expr === checkStr;
        });
    }

    //  Apparently not our aspect :)
    if (TP.notValid(entry)) {
        return this;
    }

    //  Adjust the aspect to the one the entry's actually bound to. IOW, we
    //  might have bound our "value" or "data" slot to fooBarBaz... we need to
    //  see if our value for "value" or "data" is changing.
    aspect = entry.last().last().at('bindingAspect');

    //  Dampen notifications that we already have the value for.
    current = this.get(aspect);
    value = aSignal.at(TP.NEWVAL);
    if (TP.equal(value, current)) {
        return this;
    }

    //  If we have a fixed setter call it and presume it will handle refresh. If
    //  we're forced to call the generic setter don't make that assumption.
    funcName = 'set' + aspect.asTitleCase();
    if (TP.canInvoke(this, funcName)) {
        this[funcName](value);
        return this;
    }

    //  Our last option is to update the local value and trigger refreshUI
    //  manually.
    this.set(aspect, value);
    this.refreshUI(aspect, value);

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getBindEntryTargetValue',
function(bindEntry, aRequest) {

    /**
     * @method getBindEntryTargetValue
     * @summary Returns the resolved value for a particular binding entry.
     * @param {TP.core.Hash} bindEntry The hash of binding information for the
     *     particular bound aspect that we want to get a value for.
     * @returns {Object|String} The value of the bound aspect. Note that if the
     *     supplied binding information for this aspect has more than 1 data
     *     expressions, this will be a String with all of the values
     *     concatenated together.
     */

    var dataExprs,
        len,
        i,
        dataExpr,
        scopeVals,
        allVals,
        fullExpr,
        wholeURI,
        result,
        finalResult,
        root,
        parts;

    dataExprs = bindEntry.at('dataExprs');

    len = dataExprs.getSize();
    if (len < 1) {
        return;
    }

    scopeVals = this.getBindingScopeValues(aRequest);

    for (i = 0; i < len; i++) {

        dataExpr = TP.trim(dataExprs.at(i));
        if (TP.isEmpty(dataExpr)) {
            continue;
        }

        //  If there are scope values use them as part of the path, otherwise
        //  just work from the data expression itself.
        if (TP.notEmpty(scopeVals)) {
            allVals = scopeVals.concat(TP.trim(dataExpr));
            fullExpr = TP.uriJoinFragments.apply(TP, allVals);
        } else {
            fullExpr = TP.trim(dataExpr);
        }

        //  Start by checking for a URI, that's the preferred ref in a bind.
        if (TP.isURIString(fullExpr)) {
            wholeURI = TP.uc(fullExpr);
        }

        if (TP.isURI(wholeURI)) {
            //  NOTE: we specifically tell the URI to *not* signal change if it
            //  has to fetch new content.
            result = wholeURI.getContent(TP.request('signalChange', false));
        } else {
            //  Not a URI... reconstruct and try using as a dotted path
            if (TP.notEmpty(scopeVals)) {
                fullExpr = allVals.join('.');
            }

            parts = fullExpr.split('.');

            root = this.resolveBindID(parts.shift());
            if (TP.notDefined(root)) {
                break;
            }

            if (parts.length !== 0) {
                result = root.get(parts.join('.'));
            } else {
                result = root;
            }
        }

        //  TODO:   what's the story with concat final value here? Should we be
        //  looking at this.isSingleValued and choosing based on that?
        if (TP.isValid(result)) {
            if (TP.isValid(finalResult)) {
                finalResult = TP.str(finalResult) + result;
            } else {
                finalResult = result;
            }
        }
    }

    return finalResult;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('resolveBuiltinBindID',
function(anID) {

    /**
     * @method resolveBuiltinBindID
     * @summary Returns the object referenced by a built-in bind ID such as
     *     $APP, $TYPE, etc. NOTE that for runtime binding variables such as
     *     $INDEX, $ODD, etc. this method returns TP.NO_RESULT as a way of
     *     indicating that there's no resolution possible.
     * @param {String} anID The binding ID to resolve.
     * @returns {Object} The referenced object, if found.
     */

    switch (anID) {
        //  system roots
        case 'TP':
            return TP;
        case 'APP':
            return APP;

        //  tag processing references (element, document, type)
        case '$SOURCE':
        case '$INST':
        case '$TAG':
            return this;
        case '$TARGET':
            return this.getDocument();
        case '$TYPE':
            return this.getType();

        //  application and app tag shortcuts
        case '$APP':
            return TP.sys.getApplication();
        case '$app':
            return TP.sys.getAppElement();

        //  runtime resolution variables. no observable IDs.
        case '$REQUEST':
        case '$FOCUS':
        case '$SELECTION':
        case '$_':
        case '$#':
        case '$INPUT':
        case '$INDEX':
        case '$FIRST':
        case '$MIDDLE':
        case '$LAST':
        case '$ODD':
        case '$EVEN':
            return TP.NO_RESULT;
        default:
            return;
    }
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('resolveBindID',
function(anID) {

    /**
     * @method resolveBindID
     * @summary Attempts to locate the object referenced by a bind ID. This
     *     method is used to resolve bind:context and other bind variables.
     * @param {String} anID The binding ID to resolve.
     * @returns {Object} The referenced object, if found.
     */

    var id,
        parent;

    //  TODO: deal with $ variables
    /*
    if (anID.charAt(0) === '$') {
        return this.resolveBindVariable(anID);
    }
    */

    //  Check whether this element is a named bind:context which matches.
    if (this.hasAttribute('bind:context')) {
        id = this.getAttribute('bind:context').trim();
        if (anID === id) {
            return this;
        }
    }

    //  Giving an element an ID will automatically treat that ID as a
    //  context reference. In the bind itself we support # or no #.
    if (this.hasAttribute('id')) {
        id = this.getAttribute('id').trim();
        if (id === anID) {
            return this;
        } else if (id === anID.replace(/^#/, '')) {
            return this;
        }
    }

    //  Simple way to bind to an enclosing tag (e.g. ux:list)
    if (TP.regex.HAS_COLON.test(anID)) {
        id = this.getTagName();
        if (id === anID) {
            return this;
        }
    }

    //  Work up the chain checking for any context references. NOTE that
    //  this avoids any global or "non-enclosing" context references which
    //  allows multiple independent trees to reuse a context name.
    parent = this.getParentNode();
    if (TP.isValid(parent) && TP.canInvoke(parent, 'resolveBindID')) {
        return parent.resolveBindID(anID);
    }

    return;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('computeBindOrigin',
function(anID) {

    /**
     * @method computeBindOriginID
     * @summary Given a potentially dotted bind path (e.g. $FOO.bar.baz)
     *     find the root object being referenced, if any, and get its ID.
     * @param {String} anID The binding ID to resolve.
     * @returns {Object} The referenced object, if found.
     */

    var root,
        obj;

    root = anID.split('.').first();

    obj = this.resolveBuiltinBindID(root);

    //  certain built-ins aren't meant to be observed. They can be resolved
    //  at runtime but they don't signal ($_, $INDEX, $ODD, $EVEN, etc). In
    //  that case the "object" returned is TP.NO_RESULT.
    if (obj === TP.NO_RESULT) {
        return;
    }

    if (TP.notValid(obj)) {
        //  resolve the bind ID if at all possible.
        obj = this.resolveBindID(root);
    }

    return obj;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getComputedAspect',
function(aspectName, aRequest) {

    /**
     * @method getComputedAspect
     * @summary Returns the value for an aspect of the receiver. The lookup
     *     proceeds from checking for bound data, then an explicit attribute
     *     value, then an instance variable.
     * @param {String} aspectName The aspect name to resolve a value for.
     * @param {TP.sig.Request|TP.core.Hash}
     * @returns {Object} The value for the aspect.
     */

    var result;

    if (this.isInboundAspect(aspectName)) {
        result = this.getInboundAspect(aspectName, aRequest);
    } else if (this.hasAttribute(aspectName)) {
        result = this.getAttribute(aspectName);
    } else {
        result = this.get(aspectName);
    }

    return TP.val(result);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getElementSource',
function(retainChildren) {

    /**
     * @method getElementSource
     * @summary Returns a string representation of the element which is a
     *     rough approximation of the original source text form.
     * @param {Boolean} [retainChildren=false] True to retain child text.
     * @returns {String} The element source string.
     */

    var tag,
        str;

    tag = this.clone(retainChildren);

    str = tag.getOuterContent();
    str = str.replace(/xmlns:(?:.*)="(?:.*)"/g, '');

    return str;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getExpansionDictionary',
function(aRequest) {

    /**
     * @method getExpansionDictionary
     * @summary Returns a dictionary of common variables used in templated
     *     transforms such as $TAG, $TARGET, etc.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {TP.core.Hash} A dictionary of templating variables.
     */

    return this.getType().getExpansionDictionary(this, TP.request(aRequest));
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getInboundAspect',
function(aspectName, aRequest) {

    /**
     * @method getInboundAspect
     * @summary Returns the bound value of the aspect on the receiver.
     * @param {String} aspectName The name of the aspect that the caller wants
     *     the bound value for.
     * @returns {Object} The value of the bound aspect.
     */

    var bindEntry;

    bindEntry = this.getInboundAspectEntry(aspectName);
    if (TP.notValid(bindEntry)) {
        return null;
    }

    return this.getBindEntryTargetValue(bindEntry, aRequest);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getInboundAspectEntries',
function(flushCache = false) {

    /**
     * @method getInboundAspectEntries
     * @summary Gets the 'inbound' bind info for all aspects on the receiver.
     * @param {Boolean} [flushCache=false] Whether or not to flush any currently
     *     cached binding info that the aspect may have generated.
     * @returns {TP.core.Hash} A hash of binding information keyed by aspect name.
     */

    var entries;

    if (TP.isFalse(flushCache)) {
        entries = this.$get('$inboundBindEntries');
        if (TP.isValid(entries)) {
            return entries;
        }
    }

    entries = this.callNextMethod();

    this.$set('$inboundBindEntries', entries);

    return entries;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getOutboundAspectEntries',
function(flushCache = false) {

    /**
     * @method getOutboundAspectEntries
     * @summary Gets the 'outbound' bind info for all aspects on the receiver.
     * @param {Boolean} [flushCache=false] Whether or not to flush any currently
     *     cached binding info that the aspect may have generated.
     * @returns {TP.core.Hash} A hash of binding information keyed by aspect
     *     name.
     */

    var entries;

    if (TP.isFalse(flushCache)) {
        entries = this.$get('$outboundBindEntries');
        if (TP.isValid(entries)) {
            return entries;
        }
    }

    entries = this.callNextMethod();

    this.$set('$outboundBindEntries', entries);

    return entries;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('getValue',
function() {

    /**
     * @method getValue
     * @summary Returns the value of the receiver.
     * @returns {String} The value in string form.
     */

    return this.getAttribute('value');
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('ignoreBindings',
function() {

    var thisref,
        bindings;

    thisref = this;

    bindings = this.get('$bindings');
    if (TP.isEmpty(bindings)) {
        return;
    }

    bindings.getKeys().forEach(function(sourceID) {
        //  TODO:   in theory we could also shouldSignalChange(false) but
        //          its tricky to know if that's valid.
        thisref.ignore(sourceID, 'FacetChange');
    });

    //  Clear the list
    this.$set('$bindings', null);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('isScalarValued',
function(aspectName) {

    /**
     * @method isScalarValued
     * @summary Returns true if the receiver deals with scalar values.
     * @description See the TP.dom.Node's 'isScalarValued()' instance method
     *     for more information.
     * @param {String} [aspectName] An optional aspect name that is being used
     *     by the caller to determine whether the receiver is scalar valued for.
     * @returns {Boolean} The receiver's scalar-valued status.
     */

    return true;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('isSingleValued',
function(aspectName) {

    /**
     * @method isSingleValued
     * @summary Returns true if the receiver deals with single values.
     * @description See the TP.dom.Node's 'isSingleValued()' instance method
     *     for more information.
     * @param {String} [aspectName] An optional aspect name that is being used
     *     by the caller to determine whether the receiver is single valued for.
     * @returns {Boolean} The receiver's single-valued status.
     */

    return true;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('normalizeData',
function(aspectName, aContentObject, aRequest) {

    /**
     * @method normalizeData
     * @summary Normalizes the data that will be used by the setData() method
     *     to set the content of the receiver.
     * @param {String} aspectName The aspect name on the receiver that the data
     *     is being normalized for.
     * @param {Object} aContentObject An object to use for content.
     * @param {TP.sig.Request} aRequest A request containing control parameters.
     */

    return aContentObject;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('normalizeValue',
function(aspectName, aContentObject, aRequest) {

    /**
     * @method normalizeValue
     * @summary Normalize the value that will be used by the setValue() method
     *     to set the content of the receiver.
     * @param {String} aspectName The aspect name on the receiver that the value
     *     is being normalized for.
     * @param {Object} aContentObject An object to use for content.
     * @param {TP.sig.Request} aRequest A request containing control parameters.
     */

    return TP.val(aContentObject);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('observeBindings',
function() {

    var thisref,
        bindings,
        entries;

    thisref = this;

    //  If we're going through a new attach/observe cycle the DOM node has been
    //  rebuilt. Flush any observes and rebuild, they may have changed.
    bindings = this.get('$bindings');
    if (TP.notEmpty(bindings)) {
        this.ignoreBindings();
    }

    bindings = TP.hc();
    this.set('$bindings', bindings);

    entries = this.getInboundAspectEntries();

    entries.perform(function(kvPair) {
        var parts,
            expr,
            uri,
            obj,
            root,
            sourceID,
            aspect,
            info;

        expr = kvPair.last().at('fullExpr');

        if (TP.isURIString(expr)) {
            uri = TP.uc(expr);
        }

        if (TP.isURI(uri)) {
            //  Don't include any fragment portions for URIs we may observe. The
            //  fragment is roughly equivalent to the aspect path we'd use to
            //  get/set the value in a dotted access path.
            parts = expr.split('#');
            sourceID = parts.shift();
            aspect = parts.join('#');
        } else {
            //  For dotted paths (or simple paths) work with the root segment.
            //  The root will be a root object ref such as TP or APP, or a
            //  variable reference $APP, $INST, etc. we need to resolve to a
            //  concrete global ID.
            parts = expr.split('.');
            root = parts.shift();
            aspect = parts.join('.');

            obj = thisref.computeBindOrigin(root);
            if (TP.isValid(obj)) {
                sourceID = TP.gid(obj);
                if (TP.canInvoke(obj, 'shouldSignalChange')) {
                    obj.shouldSignalChange(true);
                }
            } else {
                sourceID = root;
            }
        }

        //  Don't observe ourselves. Presumably we'll just resolve it at
        //  runtime but watching ourselves for change is a bad idea :).
        if (sourceID === TP.gid(thisref)) {
            return;
        }

        //  Tuck away the source ID's bind info for lookup in handler. NOTE that
        //  we do this as an array so we can track multiple entries from the
        //  same source ID as needed.
        info = bindings.at(sourceID);
        if (TP.notValid(info)) {
            info = [];
            bindings.atPut(sourceID, info);
        }
        info.push([aspect, kvPair]);

        //  Watch the source for any facet change notifications.
        thisref.observe(sourceID, 'FacetChange');
    });
});

//  ------------------------------------------------------------------------

/* eslint-disable require-await */

TP.ux.Element.Inst.defineMethod('rebuildUI',
async function(aRequest) {

    /**
     */

    var request;

    request = TP.request(aRequest);
    request.atPut('rebuilding', true);
    request.atPut('processingroot', false);
    request.atPut('sourceNode', this.getNativeNode());

    //  Recompile and pass true to cause result to replace current node.
    //  NOTE: this will end up async
    this.compile(request, true);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('refreshUI',
async function(anAspect, aValue) {

    /**
     */

    var funcName,
        request;

    funcName = 'refreshUI' + anAspect.asTitleCase();

    if (TP.canInvoke(this, funcName)) {
        return this[funcName](aValue);
    }

    request = TP.request(
        'refreshing', true,
        'aspect', anAspect,
        'value', aValue);

    return this.rebuildUI(request);
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('refreshUIData',
async function(aValue) {

    /**
     */

    this.rebuildUI();
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('refreshUIValue',
async function(aValue) {

    /**
     */

    this.setDisplayValue(aValue);
});

/* eslint-enable require-await */

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setBindEntryTargetValue',
async function(bindEntry, aValue) {

    /**
     * @method setBindEntryTargetValue
     * @summary Sets the bound aspect of the receiver to the supplied value.
     * @param {TP.core.Hash} bindEntry The hash of binding information for the
     *     particular bound aspect that we want to set a value for.
     * @param {Object} aValue The value to set onto the model.
     * @returns {TP.dom.ElementNode} The receiver.
     */

    var current,
        dataExprs,
        dataExpr,
        scopeVals,

        i,
        len,
        allVals,
        fullExpr,

        wholeURI,
        primaryURI,

        frag,

        parts,
        root,

        result,
        newValue;

    //  Dampen changes to outbound bind targets.
    current = this.getBindEntryTargetValue(bindEntry);
    if (TP.equal(current, aValue)) {
        return;
    }

    dataExprs = bindEntry.at('dataExprs');

    len = dataExprs.getSize();
    if (len < 1) {
        return;
    }

    scopeVals = this.getBindingScopeValues();

    for (i = 0; i < len; i++) {

        dataExpr = TP.trim(dataExprs.at(i));
        if (TP.isEmpty(dataExpr)) {
            continue;
        }

        //  If there are scope values use them as part of the path, otherwise
        //  just work from the data expression itself.
        if (TP.notEmpty(scopeVals)) {
            allVals = scopeVals.concat(TP.trim(dataExpr));
            fullExpr = TP.uriJoinFragments.apply(TP, allVals);
        } else {
            fullExpr = TP.trim(dataExpr);
        }

        //  Start by checking for a URI, that's the preferred ref in a bind.
        if (TP.isURIString(fullExpr)) {
            wholeURI = TP.uc(fullExpr);
        }

        if (TP.isURI(wholeURI)) {

            primaryURI = wholeURI.getPrimaryURI();
            frag = wholeURI.getFragment();

            //  Grab the result from the 'primary URI'. If the value can't
            //  be retrieved, then create a TP.lang.ValueHolder object and
            //  set its 'value' value to the value we're trying to set. Then set
            //  that as the 'whole resource' of the primary URI.

            //  Note we specifically tell the URI to *not* signal change if it
            //  has to fetch new content.
            result = await primaryURI.getContent(
                    TP.request('signalChange', false));
            if (TP.notValid(result)) {

                newValue = TP.lang.ValueHolder.construct(aValue);

                //  NB: This will mark the primaryURI as dirty.
                await primaryURI.setResource(
                    newValue, TP.hc('observeResource', true,
                                    'signalChange', true));
            } else {

                //  If no fragment could be computed, then we set the 'whole
                //  value'.
                if (TP.isEmpty(frag)) {
                    await result.set('value', aValue);
                } else {
                    //  If we got a Window wrapper as the result from the
                    //  primary URI, then we're a 'direct to GUI' binding.
                    //  Use the *whole* URI to get a reference to the
                    //  (wrapped) element and set its value.
                    //  Note here how we specifically tell the URI to *not*
                    //  signal change if it has to fetch new content.
                    if (TP.isKindOf(result, TP.core.Window)) {
                        result = await wholeURI.getContent(
                                    TP.request('signalChange', false));
                        await result.set('value', aValue);
                    } else {
                        await result.set(TP.apc(frag), aValue);
                    }
                }
            }

        } else {

            //  Not a URI... reconstruct and try using as a dotted path
            if (TP.notEmpty(scopeVals)) {
                fullExpr = allVals.join('.');
            }

            parts = fullExpr.split('.');

            root = this.resolveBindID(parts.shift());
            if (TP.notDefined(root)) {
                break;
            }

            if (parts.length !== 0) {
                result = await root.set(parts.join('.'), aValue);
            } else {
                result = root;
            }
        }
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setBoundAspect',
async function(anAspect, aValue, shouldForce) {

    /**
     * @method setBoundAspect
     * @summary Sets the value of a bound aspect. When an aspect is bound and
     *     its local value changes this method can be used to update the source
     *     object for the binding.
     * @param {String} aspectName The name of the aspect that the caller wants
     *     to set the bound value for.
     * @param {Object} aValue The value to set on the target object.
     * @returns {TP.dom.ElementNode} The receiver.
     */

    if (this.isOutboundAspect(anAspect)) {
        await this.setOutboundAspect(anAspect, aValue);
    } else if (this.isInboundAspect(anAspect)) {
        await this.setInboundAspect(anAspect, aValue, shouldForce);
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setDisplayValue',
function(aValue) {

    /**
     * @method setDisplayValue
     * @summary Sets the receiver's "value" whatever that means for this type.
     *     The two default operations here are to set a 'value' attribute and a
     *     'value' slot if possible.
     * @param {String} aValue The new value to set.
     * @returns {TP.html.Element} The receiver.
     */

    this.setAttribute('value', aValue);

    const node = this.getNativeNode();

    try {
        if (TP.notEmpty(node.value)) {
            node.value = aValue;
        } else {
            TP.nodeSetContent(node, aValue);
        }
    } catch (e) {
        TP.nodeSetContent(node, aValue);
    }

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setInboundAspect',
async function(aspectName, aValue, shouldForce) {

    /**
     * @method setInboundAspect
     * @summary Sets the value of the 'inbound aspect' named aspectName. Inbound
     *     aspects are those bound to the receiver using either 'bind:in' or
     *     'bind:io'. Note that if the binding here was a bind:in this method
     *     will warn (since setting a value should have implied a bind:io).
     * @param {String} aspectName The name of the aspect that the caller wants
     *     to set the bound value for.
     * @param {Object} aValue The value to set onto the model.
     * @returns {TP.dom.ElementNode} The receiver.
     */

    var bindEntry;

    bindEntry = this.getInboundAspectEntry(aspectName);
    if (TP.notValid(bindEntry)) {
        return this;
    }

    //  If not forcing make sure it's also an outbound aspect.
    if (TP.notTrue(shouldForce)) {
        if (!this.isOutboundAspect(aspectName)) {
            TP.ifWarn() ? TP.warn(
                `Ignored setInboundAspect of ${aspectName} wo/force.`) : 0;
            return this;
        }
    }

    await this.setBindEntryTargetValue(bindEntry, aValue);

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setOutboundAspect',
async function(aspectName, aValue) {

    /**
     * @method setOutboundAspect
     * @summary Sets the value of the 'outbound aspect' named aspectName.
     *     Outbound aspects are those bound to the receiver using either
     *     'bind:io' or 'bind:out'.
     * @param {String} aspectName The name of the aspect that the caller wants
     *     to set the bound value for.
     * @param {Object} aValue The value to set onto the model.
     * @returns {TP.dom.ElementNode} The receiver.
     */

    var bindEntry;

    bindEntry = this.getOutboundAspectEntry(aspectName);
    if (TP.notValid(bindEntry)) {
        return this;
    }

    await this.setBindEntryTargetValue(bindEntry, aValue);

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setup',
function() {

    /**
     * @method setup
     * @summary Perform the initial setup for the receiver.
     * @returns {TP.ux.Element} The receiver.
     */

    return this;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('setValue',
function(aValue, shouldSignal) {

    /**
     * @method setValue
     * @summary Sets the value of the receiver's node. For a UI element this
     *     method will ensure any display formatters are invoked. NOTE that this
     *     method does not update the receiver's bound value if it's a bound
     *     control. In fact, this method is used in response to a change in the
     *     bound value to update the display value, so this method should avoid
     *     changes to the bound value to avoid recursions.
     * @param {Object} aValue The value to set the 'value' of the node to.
     * @param {Boolean} shouldSignal Should changes be notified. If false
     *     changes are not signaled. Defaults to this.shouldSignalChange().
     * @returns {Boolean} Whether or not the value was changed from the value it
     *     had before this method was called.
     */

    var oldValue,
        newValue,
        flag;

    oldValue = this.getValue();
    newValue = this.normalizeValue('value', aValue);

    //  If the values are equal, there's nothing to do here - bail out.
    if (TP.equal(oldValue, newValue)) {
        return false;
    }

    //  Capture this new value before we update any binding or we'll recurse.
    this.$set('value', newValue, false);

    if (this.isOutboundAspect('value')) {
        //  If the element is bound, then update its bound value.
        this.setBoundAspect('value', newValue);
    }

    //  We changed, so trigger a UI refresh.
    this.refreshUI('value', newValue);

    //  NB: Use this construct this way for better performance
    if (TP.notValid(flag = shouldSignal)) {
        flag = this.shouldSignalChange();
    }

    if (flag) {
        this.$changed('value', TP.UPDATE,
            TP.hc(TP.OLDVAL, oldValue, TP.NEWVAL, newValue));
    }

    return true;
});

//  ------------------------------------------------------------------------

TP.ux.Element.Inst.defineMethod('teardown',
function() {

    /**
     * @method teardown
     * @summary Perform the final teardown for the receiver.
     * @returns {TP.ux.Element} The receiver.
     */

    return this;
});

//  ========================================================================
//  TP.ux.ActionTag
//  ========================================================================

/**
 * @type {TP.ux.ActionTag}
 */

//  ------------------------------------------------------------------------

TP.tag.ActionTag.defineSubtype('TP.ux.ActionTag');

TP.ux.ActionTag.addTraitTypes(TP.ux.Element);

//  ========================================================================
//  TP.ux.ComputedTag
//  ========================================================================

/**
 * @type {TP.ux.ComputedTag}
 * @summary A tag type that has its structure computed and also has the common
 *     aspect of all UX tags.
 */

//  ------------------------------------------------------------------------

TP.tag.ComputedTag.defineSubtype('TP.ux.ComputedTag');

TP.ux.ComputedTag.addTraitTypes(TP.ux.Element);

//  ========================================================================
//  TP.ux.InfoTag
//  ========================================================================

/**
 * @type {TP.ux.InfoTag}
 */

//  ------------------------------------------------------------------------

TP.tag.InfoTag.defineSubtype('TP.ux.InfoTag');

TP.ux.InfoTag.addTraitTypes(TP.ux.Element);

//  ========================================================================
//  TP.ux.TemplatedTag
//  ========================================================================

/**
 * @type {TP.ux.TemplatedTag}
 * @summary A tag type that is templated and also has the common aspect of all
 *     UX tags.
 */

//  ------------------------------------------------------------------------

TP.tag.TemplatedTag.defineSubtype('TP.ux.TemplatedTag');

TP.ux.TemplatedTag.addTraitTypes(TP.ux.Element);

//  ------------------------------------------------------------------------

TP.ux.TemplatedTag.Type.defineMethod('tagCompile',
async function(aRequest) {

    /**
     * @method tagCompile
     * @summary
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     */

    var elem,

        uri,
        resp,

        str,

        mime,
        nsInfo,
        nsURI,

        transformer;

    //  Make sure that we have an element to work from.
    if (!TP.isElement(elem = aRequest.at('node'))) {
        return this.raise('TP.sig.InvalidNode');
    }

    //  If we already have a compiled template we don't redo this... unless
    //  we're reacting to an update to the template source itself.
    if (TP.isValid(this.getExpansionTransform()) &&
        !this.$get('$isRecasting')) {
        return elem;
    }

    uri = this.getResourceURI(
            'template',
            TP.elementGetAttribute(elem, 'tibet:mime', true));

    if (TP.isValid(uri)) {

        //  Grab the receiver's content for processing. We do this synchronously
        //  here and we also get it in string form so we can process the markup
        //  and add default namespace as needed to make authoring more
        //  convenient.
        resp = await uri.getResource(
                TP.hc('resultType', TP.TEXT, 'signalChange', false));

        str = resp.get('result');

        mime = uri.getMIMEType();
        nsInfo = TP.w3.Xmlns.fromMIMEType(mime);
        if (TP.isValid(nsInfo)) {
            nsURI = nsInfo.at('uri');
        }
        switch (mime) {
            case TP.ietf.mime.SVG:
                //  fallthrough
            case TP.ietf.mime.XML:
                //  fallthrough
            case TP.ietf.mime.XHTML:
                //  Verify it's well-formed. If not then we already have a
                //  problem. Without a transformer we'll end up down below using
                //  the
                if (TP.isElement(TP.elementFromString(str, nsURI, true))) {
                    transformer = str.compile(uri.getVirtualLocation(), true);
                }
                break;
            default:
                TP.ifError() ?
                    TP.warn('Unknown/unsupported template MIME type ' +
                        mime + ' for ' + uri.getVirtualLocation()) : 0;
                break;
        }
    } else {
        str = this.getTypeName() + ' missing template uri';
        TP.ifError() ? TP.error(str) : 0;
        transformer = function(aDataSource, aParamHash) {
            return str;
        };
    }

    if (TP.notValid(transformer)) {
        str = this.getExpandedErrorSource(TP.wrap(elem), aRequest);

        transformer = function(aDataSource, aParamHash) {
            return str;
        };
    }

    //  Store the transform. It'll be leveraged during tagExpand.
    this.set('expansionTransform', transformer);

    return elem;
});

//  ------------------------------------------------------------------------

TP.ux.TemplatedTag.Type.defineMethod('getExpandedSource',
function(tpElement, aRequest) {

    /**
     * @method getExpandedSource
     * @summary Executes the receiver's transform (which should have been
     *     created during tagCompile) to produce the receiver's expanded source
     *     string. The getExpansionTransform method provides the transform.
     * @param {TP.dom.Element} tpElement A wrapped instance to be expanded.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {String|Array} The string or ordered pair result from expansion.
     */

    var transform,
        options;

    transform = this.getExpansionTransform(tpElement, aRequest);

    if (TP.notValid(transform)) {
        return this.getExpandedErrorSource(tpElement, aRequest);
    }

    options = this.getExpansionDictionary(tpElement, aRequest);

    try {
        return transform(tpElement, options);
    } catch (e) {
        console.trace(e);

        //  TODO:   replace with something like an "error link" / overlay??
        return this.getExpandedErrorSource(tpElement, aRequest);
    }
});

//  ------------------------------------------------------------------------

TP.ux.TemplatedTag.Type.defineMethod('getExpandedErrorSource',
function(tpElement, aRequest) {

    /**
     * @method getExpandedErrorSource
     * @summary Returns a string to be produced when the receiver's template was
     *     not able to be compiled properly. The string returned here will be
     *     injected into the DOM as a replacement for that template content.
     * @param {TP.dom.UIElementNode} tpElement The wrapped instance of the
     *     receiver being processed.
     * @param {TP.sig.Request} aRequest A request containing processing
     *     parameters and other data.
     * @returns {String} The content to provide for a placeholder DOM node.
     */

    var uri,
        str;

    uri = this.getResourceURI(
            'template',
            tpElement.getAttribute('tibet:mime', true));

    str = '<a onclick="alert(\'Repair error(s) in ' + this.getName() +
            ' template ' + uri.getVirtualLocation() + '\')" href="#" tibet:tag="' +
            tpElement.getCanonicalName() + '" tibet:no-rewrite="true">' +
            '&lt;' + tpElement.getCanonicalName() + '/&gt;' +
            '</a>';

    return str;
});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
