//  ========================================================================
/**
 * @copyright Copyright (C) 1999 Technical Pursuit Inc. (TPI) All Rights
 *     Reserved. Patents Pending, Technical Pursuit Inc. Licensed under the
 *     OSI-approved Reciprocal Public License (RPL) Version 1.5. See the RPL
 *     for your rights and responsibilities. Contact TPI to purchase optional
 *     privacy waivers if you must keep your TIBET-based source code private.
 */
//  ========================================================================

//  ------------------------------------------------------------------------
//  Driver Fixture
//  ------------------------------------------------------------------------

TP.test.GUIDriver.Type.defineMethod('getTestFixture',
function(options) {

    var testCase;

    if (TP.isValid(testCase = options.at('testCase'))) {
        return testCase.getDriver();
    }

    return testCase;
});

//  ========================================================================
//  Test Suite
//  ========================================================================

TP.test.GUIDriver.Inst.describe('Basic driver tests',
function() {

    var driver,
        windowContext,

        loadURI,
        unloadURI;

    unloadURI = TP.uc(TP.sys.cfg('path.blank_page'));

    //  ---

    this.before(
        async function(suite, options) {
            driver = this.getDriver();
            windowContext = driver.get('windowContext');

            loadURI = TP.uc('~lib_test/src/tibet/driver/GUIDriver1.xhtml');

            await driver.setLocation(loadURI);

            this.startTrackingSignals();
        });

    //  ---

    this.after(
        async function(suite, options) {

            await this.stopTrackingSignals();

            //  Unload the current page by setting it to the blank
            await driver.setLocation(unloadURI);

            //  Unregister the URI to avoid a memory leak
            loadURI.unregister();
        });

    //  ---

    this.afterEach(
        function(test, options) {
            this.getSuite().resetSignalTracking();
        });

    //  ---

    this.it('all sequences - sending events - explicit target', async function(test, options) {

        var inputTextField,

            bodyElem,
            focusedElem,

            seq;

        //  The window-level focus handlers in TIBET will focus the body
        //  here.
        windowContext.getNativeWindow().focus();

        inputTextField = TP.byId('input_text_field', windowContext, false);

        //  The first focused element in this file will be the <body>
        //  Test that theory.
        bodyElem = TP.documentGetBody(windowContext.getNativeDocument());
        focusedElem = driver.getFocusedElement();

        test.assert.isIdenticalTo(focusedElem, bodyElem);

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.LEFT, inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMClick');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.doubleClick(TP.LEFT, inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMDblClick');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.mouseDown(TP.LEFT, inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMMouseDown');
        await test.refute.didSignal(inputTextField, 'TP.sig.DOMMouseUp');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.mouseUp(TP.LEFT, inputTextField);
        await seq.run();

        await test.refute.didSignal(inputTextField, 'TP.sig.DOMMouseDown');
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMMouseUp');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.RIGHT, inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMContextMenu');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.keyDown('B', inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyDown');
        await test.refute.didSignal(inputTextField, 'TP.sig.DOMKeyPress');
        await test.refute.didSignal(inputTextField, 'TP.sig.DOMKeyUp');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.keyUp('C', inputTextField);
        await seq.run();

        await test.refute.didSignal(inputTextField, 'TP.sig.DOMKeyDown');
        await test.refute.didSignal(inputTextField, 'TP.sig.DOMKeyPress');
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyUp');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('A', inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyDown');
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyPress');
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyUp');

        test.getSuite().resetSignalTracking();

        //  Clear the value of the field for other tests.
        inputTextField.value = '';

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[Shift]', inputTextField);
        seq.sendKeys('[Shift-Up]', inputTextField);
        await seq.run();

        //  NB: Modifier and special keys don't send keypress.
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyDown');
        await test.refute.didSignal(inputTextField, 'TP.sig.DOMKeyPress');
        await test.assert.didSignal(inputTextField, 'TP.sig.DOMKeyUp');

        test.getSuite().resetSignalTracking();

        //  ---

        seq = driver.constructSequence();
        seq.sendEvent(TP.hc('type', 'change'), inputTextField);
        await seq.run();

        await test.assert.didSignal(inputTextField, 'TP.sig.DOMChange');
    });

    //  ---

    this.it('all sequences - sending events - focused target', async function(test, options) {

        var textareaField,

            focusedElem,

            seq;

        //  The window-level focus handlers in TIBET will focus the body
        //  here.
        windowContext.getNativeWindow().focus();

        textareaField = TP.byId('textarea_field', windowContext, false);

        //  Focus the textfield to make it the implicit target.
        await TP.wrap(textareaField).focus();

        //  Test to make sure that is the case.
        focusedElem = driver.getFocusedElement();

        test.assert.isIdenticalTo(focusedElem, textareaField);

        //  ---

        seq = driver.constructSequence();
        seq.click();
        await seq.run();

        await test.assert.didSignal(textareaField, 'TP.sig.DOMClick');
    });

    //  ---

    this.it('sendKeys sequence - set value', async function(test, options) {

        var inputTextField,

            seq;

        inputTextField = TP.byId('input_text_field', windowContext, false);

        seq = driver.constructSequence();
        seq.sendKeys('ABC', inputTextField);
        await seq.run();

        test.assert.isEqualTo(inputTextField.value, 'ABC');

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[ArrowLeft]', inputTextField);
        seq.sendKeys('[Backspace]', inputTextField);
        seq.sendKeys('D', inputTextField);
        await seq.run();

        test.assert.isEqualTo(inputTextField.value, 'ADC');

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('B', inputTextField);
        await seq.run();

        test.assert.isEqualTo(inputTextField.value, 'ADBC');

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[ArrowRight]', inputTextField);
        seq.sendKeys('E', inputTextField);
        await seq.run();

        test.assert.isEqualTo(inputTextField.value, 'ADBCE');

        //  ---

        //  Clear the value of the field
        inputTextField.value = '';

        seq = driver.constructSequence();
        seq.sendKeys('[Shift]abc[Shift-Up]de', inputTextField);
        await seq.run();

        test.assert.isEqualTo(inputTextField.value, 'ABCde');

        //  Clear the value of the field for other tests.
        inputTextField.value = '';
    });

    //  ---

    this.it('sendKeys sequence - tabbing to focus - explicit target', async function(test, options) {

        var inputTextField,
            textareaField,

            bodyElem,

            focusedElem,

            seq;

        //  The window-level focus handlers in TIBET will focus the body
        //  here.
        windowContext.getNativeWindow().focus();

        inputTextField = TP.byId('input_text_field', windowContext, false);
        textareaField = TP.byId('textarea_field', windowContext, false);

        //  Force the first focused element in this file to be the <body>
        bodyElem = TP.documentGetBody(windowContext.getNativeDocument());
        await TP.wrap(bodyElem).focus();

        //  Test that theory.
        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, bodyElem);

        //  ---

        //  Now force the input text field to be focused.
        await TP.wrap(inputTextField).focus();

        //  Test that theory.
        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, inputTextField);

        seq = driver.constructSequence();
        seq.sendKeys('[Tab]', inputTextField);
        await seq.run();

        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, textareaField);

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[Shift][Tab][Shift-Up]', textareaField);
        await seq.run();

        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, inputTextField);
    });

    //  ---

    this.it('sendKeys sequence - tabbing to focus - implicit target', async function(test, options) {

        var inputTextField,
            textareaField,

            bodyElem,

            focusedElem,

            seq;

        //  The window-level focus handlers in TIBET will focus the body
        //  here.
        windowContext.getNativeWindow().focus();

        inputTextField = TP.byId('input_text_field', windowContext, false);
        textareaField = TP.byId('textarea_field', windowContext, false);

        //  Force the first focused element in this file to be the <body>
        bodyElem = TP.documentGetBody(windowContext.getNativeDocument());
        await TP.wrap(bodyElem).focus();

        //  Test that theory.
        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, bodyElem);

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[Tab]');
        await seq.run();

        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, inputTextField);

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[Tab]');
        await seq.run();

        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, textareaField);

        //  ---

        seq = driver.constructSequence();
        seq.sendKeys('[Shift][Tab][Shift-Up]');
        await seq.run();

        focusedElem = driver.getFocusedElement();
        test.assert.isIdenticalTo(focusedElem, inputTextField);
    });

    //  ---

    this.it('click sequence - set value', async function(test, options) {

        var inputRadio1,
            inputRadio2,
            inputRadio3,

            val,

            seq,

            inputCheckbox1,
            inputCheckbox2,
            inputCheckbox3;

        inputRadio1 = TP.byId('input_radio_1', windowContext, false);
        inputRadio2 = TP.byId('input_radio_2', windowContext, false);
        inputRadio3 = TP.byId('input_radio_3', windowContext, false);

        val = TP.wrap(inputRadio1).getValue();
        test.assert.isEqualTo(val, 'foo');

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.LEFT, inputRadio2);
        await seq.run();

        val = TP.wrap(inputRadio1).getValue();
        test.assert.isEqualTo(val, 'bar');

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.LEFT, inputRadio3);
        await seq.run();

        val = TP.wrap(inputRadio1).getValue();
        test.assert.isEqualTo(val, 'baz');

        //  ---

        inputCheckbox1 = TP.byId('input_checkbox_1', windowContext, false);
        inputCheckbox2 = TP.byId('input_checkbox_2', windowContext, false);
        inputCheckbox3 = TP.byId('input_checkbox_3', windowContext, false);

        val = TP.wrap(inputCheckbox1).getValue();
        test.assert.isEqualTo(val, TP.ac('foo'));

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.LEFT, inputCheckbox2);
        await seq.run();

        val = TP.wrap(inputCheckbox1).getValue();
        test.assert.isEqualTo(val, TP.ac('foo', 'bar'));

        //  ---

        seq = driver.constructSequence();
        seq.click(TP.LEFT, inputCheckbox3);
        await seq.run();

        val = TP.wrap(inputCheckbox1).getValue();
        test.assert.isEqualTo(val, TP.ac('foo', 'bar', 'baz'));
    });

});

//  ------------------------------------------------------------------------
//  end
//  ========================================================================
