(function() {
    'use strict';

    var GIT_COMMAND;

    /**
     * The name of the git executable we look for to confirm installation.
     * @type {string}
     */
    GIT_COMMAND = 'git';

    module.exports = function(make, resolve, reject) {

        var isFrozen,

            gitpath,

            shouldBumpVersion,

            file,
            json,

            result,

            electronBuilder,

            appPath,
            etcPath,
            distPath,
            configPath,

            error,

            updateURL,

            electronConfig,

            versionResult,

            appname,

            versionStampedFiles,

            shouldCommitVersionChanges,

            commitAndPushVersionChanges,

            promise;

        isFrozen = make.CLI.isFrozen();
        if (isFrozen) {
            reject('Project already frozen. Please thaw and try again.');
            return;
        }

        make.info('checking for git support...');

        gitpath = make.sh.which(GIT_COMMAND);
        if (gitpath) {
            make.info('found git...');
            gitpath = gitpath.toString();
        } else {
            make.info('git not installed');
            return;
        }

        shouldBumpVersion = false;

        make.log('Build the app');

        //  If we're doing a release, we want some sanity checks for certain
        //  properties
        if (make.options.release) {

            make.log('Building the app for release');

            file = make.CLI.expandPath('~tibet_file');
            json = require(file);
            if (!json) {
                reject('Unable to check packaging information in: ' + file);
                return;
            }

            if (make.CLI.notEmpty(json.boot.profile) &&
                json.boot.profile !== 'main@production') {
                result = make.CLI.prompt.question(
                    'Your boot profile is not set to: ' +
                    '"main@production",' +
                    ' but is: "' + json.boot.profile + '".' +
                    ' This is not recommended.' +
                    '\nTo proceed with: "' + json.boot.profile +
                    '", enter \'yes\': ');
                if (!/^y/i.test(result)) {
                    reject('TIBET build cancelled.');
                    return;
                }
            }

            if (make.CLI.notEmpty(json.boot.bundled) &&
                json.boot.bundled !== true) {
                result = make.CLI.prompt.question(
                    'Your boot.bundled flag is set to: ' +
                    json.boot.bundled + '.' +
                    ' This is not recommended.' +
                    '\nTo proceed with: ' + json.boot.bundled +
                    ', enter \'yes\': ');
                if (!/^y/i.test(result)) {
                    reject('TIBET build cancelled.');
                    return;
                }
            }

            shouldBumpVersion = true;
        }

        electronBuilder = require('electron-builder');

        appPath = make.CLI.expandPath('~app');

        etcPath = make.CLI.joinPaths(
                        appPath,
                        'etc');

        distPath = make.CLI.joinPaths(
                        appPath,
                        'dist');

        configPath = make.CLI.joinPaths(
                        etcPath,
                        'electron-builder-config.json');

        updateURL = make.CLI.cfg('deploy.electron.updateURL', null, true);

        electronConfig = require(configPath);

        //  Add a 'publish' config, even though we're not publishing in this
        //  step - that takes place in our deploy command.. This 'provider' won't
        //  actually copy files anywhere, but is necessary to generate the proper
        //  .yml files for when we actually deploy.
        electronConfig.publish = [
            {
                provider: 'generic',
                url: updateURL
            }
        ];

        //  Remove the 'dist' directory so that we're not uploading all of
        //  the versions that came before.
        error = make.sh.rm('-rf', distPath);
        if (make.sh.error()) {
            return reject('Error removing dist directory: ' + error.stderr);
        }

        //  If we're stamping the version (because we're doing a 'release'
        //  build), then call upon 'tibet version' to do that and do the proper
        //  Git commands to commit all of that.
        if (shouldBumpVersion) {

            make.log('Bumping the version number and stamping into the release');

            result = make.sh.exec('tibet version --patch --suffix final');

            //  TODO: Figure out how to reject here.
            if (result.code !== 0) {
                return reject(
                    'Version stamping cancelled. Build & release cancelled.');
            }

            versionResult = make.sh.exec('tibet version --full');
            versionResult = make.CLI.clean(versionResult.stdout).trim();

            appname = make.cfg('project.name') || make.cfg('npm.name');

            versionStampedFiles =
                make.CLI.joinPaths(appPath, 'package.json') + ' ' +
                make.CLI.joinPaths(appPath,
                                    'TIBET-INF/cfg/' + appname + '.xml ') +
                make.CLI.joinPaths(appPath, 'TIBET-INF/cfg/main.xml ') +
                make.CLI.joinPaths(appPath, 'src/templates/version.js');

            result = make.CLI.prompt.question(
                'New version stamp computed as: ' + versionResult + '.' +
                ' Commit this version stamp or proceed without committing? ' +
                '? Enter \'yes\' after inspection: ');
            if (!/^y/i.test(result)) {
                shouldCommitVersionChanges = false;
            } else {
                shouldCommitVersionChanges = true;
            }

            commitAndPushVersionChanges = function() {
                make.log('Commit the version stamp changes');

                result = make.sh.exec(
                    gitpath + ' ' + 'add ' + versionStampedFiles);

                if (result.code !== 0) {
                    return reject(
                        'Adding files for version number stamping to git ' +
                        'caused: ' + result.code);
                }

                result = make.sh.exec(
                            gitpath + ' ' +
                            'commit ' +
                            '-am ' +
                            '"Update the version to: ' + versionResult + '"');

                if (result.code !== 0) {
                    return reject('Committing version number to git caused: ' +
                                    result.code);
                }

                make.log('Push the version stamp changes');

                result = make.sh.exec(
                            gitpath + ' ' +
                            'push');

                if (result.code !== 0) {
                    return reject('Pushing version information to git caused: ' +
                                    result.code);
                }
            };
        }

        promise = new Promise(function(resolver, rejector) {

            make.log('Freezing TIBET into the release');

            make.sh.exec('tibet freeze --standalone --source');

            return electronBuilder.build({
                mac: ['zip:universal'],
                win: ['nsis:x64', 'nsis:arm64'],
                linux: ['snap:x64'],
                config: electronConfig
            }).then(
                function() {
                    make.log('Thawing TIBET to reset it for development');

                    make.sh.exec('tibet thaw --force');

                    if (shouldBumpVersion && shouldCommitVersionChanges) {
                        commitAndPushVersionChanges();
                    }

                    return resolver();
                },
                function(err) {
                    make.log('Thawing TIBET to reset it for development');

                    make.sh.exec('tibet thaw --force');

                    if (shouldBumpVersion) {
                        make.log('Release cancelled. Revert any uncommitted ' +
                                    'release oriented changes to: ' +
                                    versionStampedFiles);
                    }

                    return rejector(err);
                }
            );
        });

        promise.then(resolve, reject);
    };

    module.exports.options = {
        timeout: 1000 * 60 * 10
    };

}());
